# Copyright (c) 2012 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

"""helper functions for the functions called by bunchdt"""
import itertools
import geometry.surface

def grouper(n, iterable, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return itertools.izip_longest(fillvalue=fillvalue, *args)
    
def getcoords(dt):
    """return the coordinates of the surface"""
    n_vertices_index = dt.objls.index('Number_of_Vertices')
    first_x = n_vertices_index + 1 # X of first coordinate
    pts = dt.obj[first_x:]
    return list(grouper(3, pts))

def area(dt):
    """area of the surface"""
    coords = getcoords(dt)
    return geometry.surface.area(coords)
    
def height(dt):
    """height of the surface"""
    coords = getcoords(dt)
    return geometry.surface.height(coords)
    
def width(dt):
    """width of the surface"""
    coords = getcoords(dt)
    return geometry.surface.width(coords)
    
def azimuth(dt):
    """azimuth of the surface"""
    coords = getcoords(dt)
    return geometry.surface.azimuth(coords)
    
def tilt(dt):
    """tilt of the surface"""
    coords = getcoords(dt)
    return geometry.surface.tilt(coords)
    



# EPlusInterface (EPI) - An interface for EnergyPlus
# Copyright (C) 2004 Santosh Philip

# This file is part of EPlusInterface.
# 
# EPlusInterface is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# EPlusInterface is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with EPlusInterface; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 


# Santosh Philip, the author of EPlusInterface, can be contacted at the following email address:
# santosh_philip AT yahoo DOT com
# Please send all bug reports, enhancement proposals, questions and comments to that address.
# 
# VERSION: 0.001



import string


def writeStr2File(pathname,s):
#	writes a string to file
	fname=pathname
	f=open(fname,'wb')
	f.write(s)
	f.close()
	
def readfile(pathname):
#	retrun the data in the file
	f=open(pathname,'rb')
	data=f.read()
	f.close()	
	return data




# import sys
# pathnameto_eppy = '../'
# sys.path.append(pathnameto_eppy)
# from eppy import modeleditor
# from eppy.modeleditor import IDF
# iddfile = "./resources/iddfiles/Energy+V7_2_0.idd"
# fname1 = "./resources/idffiles/V_7_2/smallfile.idf"
# IDF.setiddname(iddfile)
# idf1 = IDF(fname1)
# idf1.newidfobject('BUILDINGSURFACE:DETAILED')
# idf1.newidfobject('FENESTRATIONSURFACE:DETAILED')
# idf1.newidfobject('WALL:EXTERIOR')
# bsd = idf1.newidfobject('BUILDINGSURFACE:DETAILED')
# bsd = idf1.newidfobject('BUILDINGSURFACE:DETAILED')
# print bsd
# fname1 = "./resources/idffiles/V_7_0/5ZoneSupRetPlenRAB.idf"
# idf1 = IDF(fname1)
# surfaces = idf1.idfobjects['BUILDINGSURFACE:DETAILED']
# surface = surfaces[0]
# print surface.__functions
# # print "surface azimuth =",  surface.azimuth, "degrees"
# # print "surface tilt =", surface.tilt, "degrees"
# # print "surface area =", surface.area, "m2"
# # surface.width
# # surface.height

from copy import copy
import sys
pathnameto_eppy = '../'
sys.path.append(pathnameto_eppy)
from eppy import modeleditor
from eppy.modeleditor import IDF
from eppy import simplesurface
iddfile = "./resources/iddfiles/Energy+V7_2_0.idd"
fname1 = "./resources/idffiles/V_7_2/smallfile.idf"
IDF.setiddname(iddfile)
fname1 = "./resources/idffiles/V_7_0/5ZoneSupRetPlenRAB.idf"
fname1 = "/Volumes/Server/Staff/Santosh/mike_eplus/working/5ZoneSupRetPlenRAB.idf"
idf1 = IDF(fname1)

zones = idf1.idfobjects['ZONE']
for zone in zones:
    # print zone.Floor_Area, zone.Volume, zone.Ceiling_Height
    try:
        area = modeleditor.zonearea(idf1, zone.Name)
        volume =  modeleditor.zonevolume(idf1, zone.Name)
        height = modeleditor.zoneheight(idf1, zone.Name)
        zone.Floor_Area = area
        zone.Volume = volume
        zone.Ceiling_Height = height
        print zone.Name, area, height, volume
    except ValueError, e:
        print zone.Name, " - messed up"
# 
# 
# 
surfaces = idf1.idfobjects['BUILDINGSURFACE:DETAILED']
csurfaces = copy(surfaces)

count = 0
print len(surfaces)
for surface in csurfaces:
    s = simplesurface.simplesufrace(idf1, surface, deletebsd=True, setto000=True)
    if not s:
        print surface.Name
    count += 1
print count
print len(csurfaces)    
print len(surfaces)    

fens = idf1.idfobjects['FENESTRATIONSURFACE:DETAILED']
cfens = copy(fens)
print len(fens)  
count = 0
for fen in cfens:
    f = simplesurface.simplefenestration(idf1, fen, deletebsd=True, setto000=True)
    if not f:
        print fen.Name
    count += 1
print count 
print len(cfens) 
print len(fens) 

newfname1 = "/Volumes/Server/Staff/Santosh/mike_eplus/working/5ZoneSupRetPlenRAB_flat.idf"

idf1.saveas(newfname1)


#!/usr/bin/env python

## EPlusInterface (EPI) - An interface for EnergyPlus
## Copyright (C) 2004 Santosh Philip
##
## This file is part of EPlusInterface.
## 
## EPlusInterface is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## EPlusInterface is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with EPlusInterface; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
##
##
## Santosh Philip, the author of EPlusInterface, can be contacted at the following email address:
## santosh_philip AT yahoo DOT com
## Please send all bug reports, enhancement proposals, questions and comments to that address.
## 
## VERSION: 0.003
## Last updated 21 Apr 2004

import sys
import getopt
import os
import string
import user

sys.path.append(user.home+'/Library/PythonFiles')
import mylib1,mylib2,mylib3
import eplus2,time

import cPickle

def nocomment(st,com):
    """
    just like the comment in python.
    removes any text after the phrase 'com'
    """
    ls=st.splitlines()
    for i in range(len(ls)):
        el=ls[i]
        pt=el.find(com)
        if pt!=-1:
            ls[i]=el[:pt]
    return '\n'.join(ls)

def get_nocom_vars(st):
    """
    input 'st' which is the Energy+.idd file as a string
    returns (st1,st2,lss)
    st1= with all the ! comments striped
    st2= strips all comments - both the '!' and '\\'
    lss= nested list of all the variables in Energy+.idd file
    """
    nocom=nocomment(st,'!')# remove '!' comments
    st1=nocom
    nocom1=nocomment(st1,'\\')# remove '\' comments
    st1=nocom
    st2=nocom1
    ls=string.split(st2,';')
    lss=[]
    
    # break the .idd file into a nested list
    #=======================================
    for el in ls:
        item=string.split(el,',')
        lss.append(item)
    for i in range(0,len(lss)):
        for j in range(0,len(lss[i])):
            lss[i][j]=lss[i][j].strip()
    if len(lss)>1:lss.pop(-1)
    #=======================================    
    
    #st1 has the '\' comments --- looks like I don't use this
    #lss is the .idd file as a nested list
    return (st1,st2,lss)

def removeblanklines(st):
    """
    removeblanklines(st)
    returns the string after
    remove blank lines in 'st'
    """
    linesep=mylib3.getlinesep(st)
    ls=st.split(linesep)
    lss=[]
    for el in ls:
        ell=el.strip()
        if ell!='':
            lss.append(el)
    st1=linesep.join(lss)
    return st1


def extractidddata(fname,debug=False):
    """
    extracts all the needed information out of the idd file
    if debug is True, it generates a series of text files.
    Each text file is incrementally different. You can do a diff 
    see what the change is
    """
    from StringIO import StringIO
    if isinstance(fname, (file, StringIO)):
        st = fname.read()
    else:
        st=mylib2.readfile(fname)
    (nocom,nocom1,blocklst)=get_nocom_vars(st)
    
    
    st=nocom
    st1=removeblanklines(st)
    if debug:
        mylib1.writeStr2File('nocom2.txt',st1)
    
    
    #find the groups and the start object of the group
    #find all the group strings
    groupls=[]
    ls=st1.splitlines()
    for el in ls:
        lss=el.split()
        if lss[0].upper()=='\\group'.upper():
            groupls.append(el)
    
    
    #find the var just after each item in groupls
    groupstart=[]
    for i in range(len(groupls)):
        ii=ls.index(groupls[i])
        groupstart.append([ls[ii],ls[ii+1]])
    
    #remove the group commentline
    for el in groupls:
        ls.remove(el)
    
    if debug:
        st1='\n'.join(ls)
        mylib1.writeStr2File('nocom3.txt',st1)
    
    #strip each line
    for i in range(len(ls)):
        ls[i]=ls[i].strip()
    
    if debug:
        st1='\n'.join(ls)
        mylib1.writeStr2File('nocom4.txt',st1)
    
    #ensure that each line is a comment or variable
    #find lines that don't start with a comment
    #if this line has a comment in it
    #   then move the comment to a new line below
    lss=[]
    for i in range(len(ls)):
        #find lines that don't start with a comment
        if ls[i][0]!='\\':
            #if this line has a comment in it
            pt=ls[i].find('\\')
            if pt!=-1:
                #then move the comment to a new line below
                lss.append(ls[i][:pt].strip())
                lss.append(ls[i][pt:].strip())
            else:
                lss.append(ls[i])
        else:
            lss.append(ls[i])
    
    
    ls=lss[:]
    if debug:
        st1='\n'.join(ls)
        mylib1.writeStr2File('nocom5.txt',st1)
    
    #need to make sure that each line has only one variable - as in WindowGlassSpectralData,
    lss=[]
    for el in ls:
        # if the line is not a comment
        if el[0]!='\\':
            #test for more than one var
            ll=el.split(',')
            if ll[-1]=='':
                tmp=ll.pop()
            for elm in ll:
                if elm[-1]==';':
                    lss.append(elm.strip())
                else:
                    lss.append((elm+',').strip())
        else:
            lss.append(el)
    
    ls_debug=ls[:] # needed for the next debug - 'nocom7.txt'
    ls=lss[:]
    if debug:
        st1='\n'.join(ls)
        mylib1.writeStr2File('nocom6.txt',st1)
    
    if debug:
        #need to make sure that each line has only one variable - as in WindowGlassSpectralData,
        #this is same as above.
        # but the variables are put in without the ';' and ','
        #so we can do a diff between 'nocom7.txt' and 'nocom8.txt'. Should be identical
        lss_debug=[]
        for el in ls_debug:
            # if the line is not a comment
            if el[0]!='\\':
                #test for more than one var
                ll=el.split(',')
                if ll[-1]=='':
                    tmp=ll.pop()
                for elm in ll:
                    if elm[-1]==';':
                        lss_debug.append(elm[:-1].strip())
                    else:
                        lss_debug.append((elm).strip())
            else:
                lss_debug.append(el)
        
        ls_debug=lss_debug[:]
        st1='\n'.join(ls_debug)
        mylib1.writeStr2File('nocom7.txt',st1)
    
    
    #replace each var with '=====var======'
    #join into a string,
    #split using '=====var====='
    for i in range(len(lss)):
        #if the line is not a comment
        if lss[i][0]!='\\':
            lss[i]='=====var====='
    
    st2='\n'.join(lss)
    lss=st2.split('=====var=====\n')
    lss.pop(0) # the above split generates an extra item at start
    
    if debug:
        fname='nocom8.txt'
        f=open(fname,'wb')
        k=0
        for i in range(len(blocklst)):
            for j in range(len(blocklst[i])):
                f.write(blocklst[i][j]+'\n')
                f.write(lss[k])
                k=k+1
        
        f.close()
    
    #map the structure of the comments -(this is 'lss' now) to 
    #the structure of blocklst - blocklst is a nested list
    #make lss a similar nested list
    k=0
    lst=[]
    for i in range(len(blocklst)):
        lst.append([])
        for j in range(len(blocklst[i])):
            lst[i].append(lss[k])
            k=k+1
    
    
    if debug:
        fname='nocom9.txt'
        f=open(fname,'wb')
        k=0
        for i in range(len(blocklst)):
            for j in range(len(blocklst[i])):
                f.write(blocklst[i][j]+'\n')
                f.write(lst[i][j])
                k=k+1
        
        f.close()
            
    
    
    #break up multiple line comment so that it is a list
    for i in range(len(lst)):
        for j in range(len(lst[i])):
            lst[i][j]=lst[i][j].splitlines()
            # remove the '\'
            for k in range(len(lst[i][j])):
                lst[i][j][k]=lst[i][j][k][1:]
    commlst=lst
    
    
    #copied with minor modifications from readidd2_2.py -- which has been erased ha !
    c=lst
    lss=[]
    for i in range(0,len(c)):
        ls=[]
        for j in range(0,len(c[i])):
            it=c[i][j]
            dt={}
            for el in it:
                if len(el.split())==0:
                    break
                dt[el.split()[0].lower()]=[]
            
            for el in it:
                if len(el.split())==0:
                    break
                dt[el.split()[0].lower()].append(string.join(el.split()[1:]))
                                
            
            ls.append(dt)
        
        lss.append(ls)
    commdct=lss
    
    return blocklst,commlst,commdct


def getobjectref(blocklst,commdct):
    """
    makes a dictionary of object-lists
    each item in the dictionary points to a list of tuples
    the tuple is (objectname, fieldindex)
    """
    objlst_dct={}
    for eli in commdct:
        for elj in eli:
            if elj.has_key('object-list'):
                objlist=elj['object-list'][0]
                objlst_dct[objlist]=[] 
    
    for objlist in objlst_dct.keys():
        for i in range(len(commdct)):
            for j in range(len(commdct[i])):
                if commdct[i][j].has_key('reference'):
                    for ref in commdct[i][j]['reference']:
                        if ref==objlist:                        
                            objlst_dct[objlist].append((blocklst[i][0],j))
    return objlst_dct

        
# blocklst,commlst,commdct=extractidddata('Energy+.idd_old')
# dct=getobjectref(blocklst,commdct)
# 
# for el in dct.keys():
#   print el
#   print dct[el]
# fname='./Energy+.idd'
# debug=True
# blocklst,commlst,commdct=extractidddata(fname,debug)

# blocklst,commlst,commdct=extractidddata('Energy+.idd_old')


#mylib2.cpickledump((blocklst,lss),'block_comm.pik')
#this file is identical to the one produced by readidd2_2.py



# Copyright (c) 2012 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

"""read the html outputs"""


import string
import collections
from bs4 import BeautifulSoup, NavigableString, Tag

class NotSimpleTable(Exception):
    pass
        

def is_simpletable(table):
    """test if the table has only strings in the cells"""
    tds = table('td')
    for td in tds:
        if td.contents != []:
            if len(td.contents) == 1:
                if not isinstance(td.contents[0], NavigableString):
                    return False
            else:
                return False
    return True

def table2matrix(table):
    """convert a table to a list of lists - a 2D matrix"""

    if not is_simpletable(table):
        raise NotSimpleTable, "Not able read a cell in the table as a string"
    rows = []
    for tr in table('tr'):
        row = []
        for td in tr('td'):
            try:
                row.append(td.contents[0])
            except IndexError, e:
                row.append('')
        rows.append(row)
    return rows

def table2val_matrix(table):
    """convert a table to a list of lists - a 2D matrix
    Converts numbers to float"""
    if not is_simpletable(table):
        raise NotSimpleTable, "Not able read a cell in the table as a string"
    rows = []
    for tr in table('tr'):
        row = []
        for td in tr('td'):
            try:
                val = td.contents[0]
            except IndexError, e:
                row.append('')
            else:
                try:
                    val = float(val)
                    row.append(val)
                except ValueError, e:
                    row.append(val)
        rows.append(row)
    return rows


def titletable(html_doc, tofloat=True):
    """return a list of [(title, table), .....]
    
    title = previous item with a <b> tag
    table = rows -> [[cell1, cell2, ..], [cell1, cell2, ..], ..]"""
    soup = BeautifulSoup(html_doc)
    btables = soup.find_all(['b', 'table']) # find all the <b> and <table> 
    titletables = []
    for i, item in enumerate(btables):
        if item.name == 'table':
            for j in range(i + 1):
                if btables[i-j].name == 'b':# step back to find a <b>
                    break
            titletables.append((btables[i-j], item))
    if tofloat:
        t2m = table2val_matrix
    else:
        t2m = table2matrix
    titlerows = [(tl.contents[0], t2m(tb)) for tl, tb in titletables]
    return titlerows
    
def _has_name(soup_obj):
    """checks if soup_obj is really a soup object or just a string
    If it has a name it is a soup object"""
    try:
        name = soup_obj.name
        if name == None:
            return False
        return True
    except AttributeError, e:
        return False    
        
def lines_table(html_doc, tofloat=True):
    """return a list of [(lines, table), .....]
    
    lines = all the significant lines before the table. 
        These are lines between this table and 
        the previous table or 'hr' tag
    table = rows -> [[cell1, cell2, ..], [cell1, cell2, ..], ..]
    
    The lines act as a description for what is in the table
    """
    soup = BeautifulSoup(html_doc)
    linestables = []
    elements = soup.p.next_elements # start after the first para
    for element in elements:
        tabletup = []
        if not _has_name(element):
            continue
        if element.name == 'table': # hit the first table
            beforetable = []
            prev_elements = element.previous_elements # walk back and get the lines
            for prev_element in prev_elements:
                if not _has_name(prev_element):
                    continue
                if prev_element.name not in ('br', None): # no lines here
                    if prev_element.name in ('table', 'hr', 'tr', 'td'): 
                        # just hit the previous table. You got all the lines
                        break
                    if prev_element.parent.name == "p":
                        # if the parent is "p", you will get it's text anyways from the parent
                        pass
                    else:
                        if prev_element.get_text(): # skip blank lines
                            beforetable.append(prev_element.get_text())
            beforetable.reverse()
            tabletup.append(beforetable)
            function_selector = {True:table2val_matrix, False:table2matrix}
            function = function_selector[tofloat]
            tabletup.append(function(element))
        if tabletup:
            linestables.append(tabletup)
    return linestables

def _asciidigits(s):
    """if s is not ascii or digit, return an '_' """
    if s not in string.ascii_letters + string.digits:
        s = '_'
    return s

def _nospace(s):
    """replace all non-ascii, non_digit or space with '_' """
    return ''.join([_asciidigits(i) for i in s])
    

def _transpose(arr):
    return map(list, zip(*arr))
    

def _make_ntgrid(grid):
    """make a named tuple grid
    
    [["",  "a b", "b c", "c d"],
     ["x y", 1,     2,     3 ],
     ["y z", 4,     5,     6 ],
     ["z z", 7,     8,     9 ],]
    will return
    ntcol(x_y=ntrow(a_b=1, b_c=2, c_d=3), 
          y_z=ntrow(a_b=4, b_c=5, c_d=6), 
          z_z=ntrow(a_b=7, b_c=8, c_d=9))"""
    hnames = [_nospace(n) for n in grid[0][1:]]
    vnames = [_nospace(row[0]) for row in grid[1:]]
    vnames_s = " ".join(vnames)
    hnames_s = " ".join(hnames)
    ntcol = collections.namedtuple('ntcol', vnames_s)
    ntrow = collections.namedtuple('ntrow', hnames_s)
    rdict = [dict(zip(hnames, row[1:])) for row in grid[1:]]  
    ntrows = [ntrow(**rdict[i]) for i, name in enumerate(vnames)]
    ntcols = ntcol(**dict(zip(vnames, ntrows)))
    return ntcols
    
def named_grid_h(grid):
    """make a horizontal named grid"""
    return _make_ntgrid(grid)
    
def named_grid_v(grid):
    """make a vertical named grid"""
    return _make_ntgrid(_transpose(grid))    


# Copyright (c) 2012 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.


"""sub class bunch in steps going from data, aliases, functions"""
# TODO : go thru with a fine tooth comb. make unit tests

import sys
import copy
from EPlusInterfaceFunctions import readidf
from bunch import *
import bunchhelpers

class BadEPFieldError(Exception):
    pass

class RangeError(Exception):
    pass
    

def somevalues(dt):
    """returns some values"""
    return dt.Name, dt.Construction_Name, dt.obj
    
def extendlist(lst, i, value=''):
    """extend the list so that you have i-th value"""
    if i < len(lst):
        pass
    else:
        lst.extend([value, ] * (i - len(lst) + 1))

class EpBunch_1(Bunch):
    """Has data in bunch"""
    def __init__(self, obj, objls, objidd, *args, **kwargs):
        super(EpBunch_1, self).__init__(*args, **kwargs)
        self.obj = obj
        self.objls = objls
        self.objidd = objidd
    def __setattr__(self, name, value):
        if name in ('obj', 'objls', 'objidd'):
            super(EpBunch_1, self).__setattr__(name, value)
            return None
        elif name in self['objls']:
            i = self['objls'].index(name)
            try:
                self['obj'][i] = value
            except IndexError, e:
                extendlist(self['obj'], i)
                self['obj'][i] = value
        else:
            s = "unable to find field %s" % (name, )
            raise BadEPFieldError(s)
    def __getattr__(self, name):
        if name in ('obj', 'objls', 'objidd'):
            return super(EpBunch_1, self).__getattr__(name)
        elif name in self['objls']:
            i = self['objls'].index(name)
            try:
                return self['obj'][i]
            except IndexError, e:
                return ''
        else:
            s = "unable to find field %s" % (name, )
            raise BadEPFieldError(s)
    def __repr__(self):
        """print this as an idf snippet"""
        lines = [str(val) for val in self.obj]
        comments = [comm.replace('_', ' ') for comm in self.objls]
        lines[0] = "%s," % (lines[0], ) # comma after first line
        for i, line in enumerate(lines[1:-1]):
            lines[i + 1] = '    %s,' % (line, ) # indent and comma
        lines[-1] = '    %s;' % (lines[-1], )# ';' after last line
        lines = [line.ljust(26) for line in lines] # ljsut the lines
        filler = '%s    !- %s'
        nlines = [filler % (line, comm) for line, 
            comm in zip(lines[1:], comments[1:])]# adds comments to line
        nlines.insert(0, lines[0])# first line without comment
        s = '\n'.join(nlines)
        return '\n%s\n' % (s, )

class EpBunch_2(EpBunch_1):
    """Has data, aliases in bunch"""
    def __init__(self, obj, objls, objidd, *args, **kwargs):
        super(EpBunch_2, self).__init__(obj, objls, objidd, *args, **kwargs)
    def __setattr__(self, name, value):
        if name == '__aliases':
            self[name] = value
            return None
        try:
            origname = self['__aliases'][name]
            super(EpBunch_2, self).__setattr__(origname, value)
        except KeyError, e:
            super(EpBunch_2, self).__setattr__(name, value)
    def __getattr__(self, name):
        if name == '__aliases':
            return super(EpBunch_2, self).__getattr__(name)
        try:
            origname = self['__aliases'][name]
            return super(EpBunch_2, self).__getattr__(origname)
        except KeyError, e:
            return super(EpBunch_2, self).__getattr__(name)
    
class EpBunch_3(EpBunch_2):
    """Has data, aliases, functions in bunch"""
    def __init__(self, obj, objls, objidd, *args, **kwargs):
        super(EpBunch_3, self).__init__(obj, objls, objidd, *args, **kwargs)
        self['__functions'] = {}
    def __setattr__(self, name, value):
        if name == '__functions':
            self[name] = value
            return None
        try:
            origname = self['__functions'][name]
            self[origname] = value
        except KeyError, e:
            super(EpBunch_3, self).__setattr__(name, value)
    def __getattr__(self, name):
        if name == '__functions':
            return self['__functions']
        try:
            func = self['__functions'][name]
            if isinstance(func, EpBunchFunctionClass):
                return func.func
            else:
                return func(self)
        except KeyError, e:
            return super(EpBunch_3, self).__getattr__(name)
            
class EpBunch_4(EpBunch_3):
    """h implements __getitem__ and __setitem__"""
    def __init__(self, obj, objls, objidd, *args, **kwargs):
        super(EpBunch_4, self).__init__(obj, objls, objidd, *args, **kwargs)
    def __getitem__(self, key):
        if key in ('obj', 'objls', 'objidd', '__functions', '__aliases'):
            return super(EpBunch_4, self).__getitem__(key)
        elif key in self['objls']:
            i = self['objls'].index(key)
            try:
                return self['obj'][i]
            except IndexError, e:
                return ''
        else:
            s = "unknown field %s" % (key, )
            raise BadEPFieldError(s)
    def __setitem__(self, key, value):
        if key in ('obj', 'objls', 'objidd', '__functions', '__aliases'):
            super(EpBunch_4, self).__setitem__(key, value)
            return None
        elif key in self['objls']:
            i = self['objls'].index(key)
            try:
                self['obj'][i] = value
            except IndexError, e:
                extendlist(self['obj'], i)
                self['obj'][i] = value
            
        else:
            s = "unknown field %s" % (key, )
            raise BadEPFieldError(s)
        

# TODO unit test    
def fieldnames(bch):
    return bch.objls
    
def fieldvalues(bch):
    return bch.obj
    
class EpBunchFunctionClass(object):
    pass
    
class GetRange(EpBunchFunctionClass):
    def __init__(self, arg):
        self.bch = arg
    def func(self, fieldname):
        """get the ranges for this field"""
        bch =self.bch
        keys = ['maximum', 'minimum', 'maximum<', 'minimum>', 'type']
        index = bch.objls.index(fieldname)
        fielddct_orig = bch.objidd[index]
        fielddct = copy.deepcopy(fielddct_orig)
        therange = {}
        for key in keys:
            therange[key] = fielddct.setdefault(key, None)
        if therange['type']:
            therange['type'] = therange['type'][0]
        if therange['type'] == 'real':
            for key in keys[:-1]:
                if therange[key]:
                    therange[key] = float(therange[key][0])
        if therange['type'] == 'integer':
            for key in keys[:-1]:
                if therange[key]:
                    therange[key] = int(therange[key][0])
        return therange    

class CheckRange(EpBunchFunctionClass):
    def __init__(self, arg):
        self.bch = arg
    # def init(self, arg):
    #     self.bch = arg
    def func(self, fieldname):
        """throw exception if the out of range"""
        bch = self.bch
        fieldvalue = bch[fieldname]
        therange = bch.getrange(fieldname)
        keys = ['maximum', 'minimum', 'maximum<', 'minimum>']
        index = bch.objls.index(fieldname)
        # fielddct = bch.objidd[index]
        if therange['maximum'] != None:
            if fieldvalue > therange['maximum']:
                s = "Value %s is not less or equal to the 'maximum' of %s"
                s = s % (fieldvalue, therange['maximum'])
                raise RangeError(s)
        if therange['minimum'] != None:
            if fieldvalue < therange['minimum']:
                s = "Value %s is not greater or equal to the 'minimum' of %s"
                s = s % (fieldvalue, therange['minimum'])
                raise RangeError(s)
        if therange['maximum<'] != None:
            if fieldvalue >= therange['maximum<']:
                s = "Value %s is not less than the 'maximum<' of %s"
                s = s % (fieldvalue, therange['maximum<'])
                raise RangeError(s)
        if  therange['minimum>'] != None:
            if fieldvalue <= therange['minimum>']:
                s = "Value %s is not greater than the 'minimum>' of %s"
                s = s % (fieldvalue, therange['minimum>'])
                raise RangeError(s)
        return fieldvalue

class EpBunch_5(EpBunch_4):
    """implements getrange, checkrange, fieldnames"""
    def __init__(self, obj, objls, objidd, *args, **kwargs):
        super(EpBunch_5, self).__init__(obj, objls, objidd, *args, **kwargs)
        self['__functions']['fieldnames'] = fieldnames
        self['__functions']['fieldvalues'] = fieldvalues
        self['__functions']['getrange'] = GetRange(self)
        self['__functions']['checkrange'] = CheckRange(self)
        
            
EpBunch = EpBunch_5

def main():

    # read code
    # iddfile = "../iddfiles/Energy+V6_0.idd"
    iddfile = "./walls.idd"
    fname = "./walls.idf" # small file with only surfaces
    data, commdct = readidf.readdatacommdct(fname, iddfile=iddfile)

    # setup code walls - can be generic for any object
    dt = data.dt
    dtls = data.dtls
    wall_i = dtls.index('BuildingSurface:Detailed'.upper())
    wallkey = 'BuildingSurface:Detailed'.upper()

    dwalls = dt[wallkey]
    dwall = dwalls[0]

    wallfields = [comm.get('field') for comm in commdct[wall_i]]
    wallfields[0] = ['key']
    wallfields = [field[0] for field in wallfields]
    wall_fields = [bunchhelpers.makefieldname(field) for field in wallfields]
    print wall_fields[:20]

    bwall = EpBunch(dwall, wall_fields)

    print bwall.Name
    print data.dt[wallkey][0][1]
    bwall.Name = 'Gumby'
    print bwall.Name
    print data.dt[wallkey][0][1]
    print

    # set aliases
    bwall.__aliases = {'Constr':'Construction_Name'} 

    print "wall.Construction_Name = %s" % (bwall.Construction_Name, )
    print "wall.Constr = %s" % (bwall.Constr, )
    print
    print "change wall.Constr"
    bwall.Constr = 'AnewConstr'
    print "wall.Constr = %s" % (bwall.Constr, )
    print "wall.Constr = %s" % (data.dt[wallkey][0][3], )
    print

    # add functions
    bwall.__functions = {'svalues':somevalues} 

    print bwall.svalues
    print bwall.__functions



if __name__ == '__main__':
    main()




# Copyright (c) 2012 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

idfsnippet = """
Building,
  Building,                !- Name
  30.,                     !- North Axis {deg}
  City,                    !- Terrain
  0.04,                    !- Loads Convergence Tolerance Value
  0.4,                     !- Temperature Convergence Tolerance Value {deltaC}
  FullExterior,            !- Solar Distribution
  25,                      !- Maximum Number of Warmup Days
  6;                       !- Minimum Number of Warmup Days

Building,
  Building,                !- Name
  30,                       !- North Axis {deg}
  City;                    !- Terrain

Building,
  BuildinG,                !- Name
  30.001,                       !- North Axis {deg}
  CitY;                    !- Terrain

Building,
  Building,                !- Name
  30.,                     !- North Axis {deg}
  City,                    !- Terrain
  0.04,                    !- Loads Convergence Tolerance Value
  0.4,                     !- Temperature Convergence Tolerance Value {deltaC}
  FullExterior,            !- Solar Distribution
  25,                      !- Maximum Number of Warmup Days
  7;                       !- Minimum Number of Warmup Days


Zone,
  PLENUM-1,                !- Name
  0,                       !- Direction of Relative North {deg}
  0,                       !- X Origin {m}
  0,                       !- Y Origin {m}
  0,                       !- Z Origin {m}
  1,                       !- Type
  1,                       !- Multiplier
  0.609600067,             !- Ceiling Height {m}
  283.2;                   !- Volume {m3}

Zone,
  SPACE1-1,                !- Name
  0,                       !- Direction of Relative North {deg}
  0,                       !- X Origin {m}
  0,                       !- Y Origin {m}
  0,                       !- Z Origin {m}
  1,                       !- Type
  1,                       !- Multiplier
  2.438400269,             !- Ceiling Height {m}
  239.247360229;           !- Volume {m3}

Zone,
  SPACE2-1,                !- Name
  0,                       !- Direction of Relative North {deg}
  0,                       !- X Origin {m}
  0,                       !- Y Origin {m}
  0,                       !- Z Origin {m}
  1,                       !- Type
  1,                       !- Multiplier
  2.438400269,             !- Ceiling Height {m}
  103.311355591;           !- Volume {m3}

Zone,
  SPACE3-1,                !- Name
  0,                       !- Direction of Relative North {deg}
  0,                       !- X Origin {m}
  0,                       !- Y Origin {m}
  0,                       !- Z Origin {m}
  1,                       !- Type
  1,                       !- Multiplier
  2.438400269,             !- Ceiling Height {m}
  239.247360229;           !- Volume {m3}

Zone,
  SPACE4-1,                !- Name
  0,                       !- Direction of Relative North {deg}
  0,                       !- X Origin {m}
  0,                       !- Y Origin {m}
  0,                       !- Z Origin {m}
  1,                       !- Type
  1,                       !- Multiplier
  2.438400269,             !- Ceiling Height {m}
  103.311355591;           !- Volume {m3}

Zone,
  SPACE5-1,                !- Name
  0,                       !- Direction of Relative North {deg}
  0,                       !- X Origin {m}
  0,                       !- Y Origin {m}
  0,                       !- Z Origin {m}
  1,                       !- Type
  1,                       !- Multiplier
  2.438400269,             !- Ceiling Height {m}
  447.682556152;           !- Volume {m3}

Zone,
  Sup-PLENUM-1,            !- Name
  0,                       !- Direction of Relative North {deg}
  0,                       !- X Origin {m}
  0,                       !- Y Origin {m}
  0,                       !- Z Origin {m}
  1,                       !- Type
  1,                       !- Multiplier
  0.45,                    !- Ceiling Height {m}
  208.6;                   !- Volume {m3}

BuildingSurface:Detailed,
    WALL-1PF,                !- Name
    WALL,                    !- Surface Type
    WALL-1,                  !- Construction Name
    PLENUM-1,                !- Zone Name
    Outdoors,                !- Outside Boundary Condition
    ,                        !- Outside Boundary Condition Object
    SunExposed,              !- Sun Exposure
    WindExposed,             !- Wind Exposure
    0.50000,                 !- View Factor to Ground
    4,                       !- Number of Vertices
    0.0,0.0,3.0,  !- X,Y,Z ==> Vertex 1 {m}
    0.0,0.0,2.4,  !- X,Y,Z ==> Vertex 2 {m}
    30.5,0.0,2.4,  !- X,Y,Z ==> Vertex 3 {m}
    30.5,0.0,3.0;  !- X,Y,Z ==> Vertex 4 {m}

"""




# Copyright (c) 2012 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

"""get the loop data to draw the diagram
Other notes:
- tested for idd version 6.0
- when E+ is updated, run versionchangecheck.py for the following objects
uses the following objects
['plantloop', ]
"""

import sys
# sys.path.append('../EPlusInputcode')
from EPlusInterfaceFunctions import readidf

def extractfields(data, commdct, objkey, fieldlists):
    """get all the objects of objkey. 
    fieldlists will have a fieldlist for each of those objects.
    return the contents of those fields"""
    # TODO : this assumes that the field list identical for 
    # each instance of the object. This is not true. 
    # So we should have a field list for each instance of the object 
    # and map them with a zip
    objindex = data.dtls.index(objkey)
    objcomm = commdct[objindex]
    objfields = []
    # get the field names of that object
    for dct in objcomm[0:]:
        try:
            thefieldcomms = dct['field']
            objfields.append(thefieldcomms[0])
        except KeyError, e:
            objfields.append(None)
    fieldindexes = []
    for fieldlist in fieldlists:
        fieldindex = []
        for item in fieldlist:
            if isinstance(item, int):
                fieldindex.append(item)
            else:
                fieldindex.append(objfields.index(item) + 0)
                # the index starts at 1, not at 0
        fieldindexes.append(fieldindex)
    theobjects = data.dt[objkey]
    fieldcontents = []
    for theobject, fieldindex in zip(theobjects, fieldindexes):
        innerlst = []
        for item in fieldindex:
            try:
                innerlst.append(theobject[item])
            except IndexError, e:
                break
        fieldcontents.append(innerlst)
        # fieldcontents.append([theobject[item] for item in fieldindex])
    return fieldcontents

def plantloopfieldlists(data):
    """return the plantloopfield list"""
    objkey = 'plantloop'.upper()
    numobjects = len(data.dt[objkey])
    return [['Name',
       'Plant Side Inlet Node Name',
       'Plant Side Outlet Node Name',
       'Plant Side Branch List Name',
       'Demand Side Inlet Node Name',
       'Demand Side Outlet Node Name',
       'Demand Side Branch List Name']] * numobjects
    
def plantloopfields(data, commdct):
    """get plantloop fields to diagram it"""
    fieldlists = plantloopfieldlists(data)
    objkey = 'plantloop'.upper()
    return extractfields(data, commdct, objkey, fieldlists)
    
def branchlist2branches(data, commdct, branchlist):
    """get branches from the branchlist"""
    objkey = 'BranchList'.upper()
    theobjects = data.dt[objkey]
    fieldlists = []
    objnames = [obj[1] for obj in theobjects]
    for theobject in theobjects:
        fieldlists.append(range(2, len(theobject)))
    blists = extractfields(data, commdct, objkey, fieldlists)
    thebranches = [branches for name, branches in zip(objnames, 
                                        blists) if name == branchlist]
    return thebranches[0]
    
def branch_inlet_outlet(data, commdct, branchname):
    """return the inlet and outlet of a branch"""
    objkey = 'Branch'.upper()
    theobjects = data.dt[objkey]
    theobject = [obj for obj in theobjects if obj[1] == branchname]
    theobject = theobject[0]
    inletindex = 6
    outletindex = len(theobject) - 2
    return [theobject[inletindex], theobject[outletindex]]
    
def splittermixerfieldlists(data, commdct, objkey):
    """docstring for splittermixerfieldlists"""
    objkey = objkey.upper()
    objindex = data.dtls.index(objkey)
    objcomms = commdct[objindex]
    theobjects = data.dt[objkey]
    fieldlists = []
    for theobject in theobjects:
        fieldlist = range(1,len(theobject))
        fieldlists.append(fieldlist)
    return fieldlists

def splitterfields(data, commdct):
    """get splitter fields to diagram it"""
    objkey = "Connector:Splitter".upper()
    fieldlists = splittermixerfieldlists(data, commdct, objkey)
    return extractfields(data, commdct, objkey, fieldlists)

def mixerfields(data, commdct):
    """get mixer fields to diagram it"""
    objkey = "Connector:Mixer".upper()
    fieldlists = splittermixerfieldlists(data, commdct, objkey)
    return extractfields(data, commdct, objkey, fieldlists)

        
def repeatingfields(theidd, commdct, objkey, flds):
    """return a list of repeating fields
    fld is in format 'Component %s Name'
    so flds = [fld % (i, ) for i in range(n)]
    does not work for 'fields as indicated' """
    # TODO : make it work for 'fields as indicated'
    if type(flds) != list:
        flds = [flds] # for backward compatability
    objindex = theidd.dtls.index(objkey)
    objcomm = commdct[objindex]
    allfields = []
    for fld in flds:
        thefields = []
        indx = 1
        for i in range(len(objcomm)):
            try:
                thefield = fld % (indx, )
                if objcomm[i]['field'][0] == thefield:
                    thefields.append(thefield)
                    indx = indx + 1
            except KeyError, e:
                pass
        allfields.append(thefields)
    allfields = zip(*allfields)
    return [item for sublist in allfields for item in sublist]

def objectcount(data, key):
    """return the count of objects of key"""
    objkey = key.upper()
    return len(data.dt[objkey])

def getfieldindex(data, commdct, objkey, fname):
    """given objkey and fieldname, return its index"""
    objindex = data.dtls.index(objkey)
    objcomm = commdct[objindex]
    for i, item in enumerate(objcomm): 
        try:
            if item['field'] == [fname]:
                break
        except KeyError, e:
            pass
    return i
    
def getadistus(data, commdct):
    """docstring for fname"""
    objkey = "ZoneHVAC:AirDistributionUnit".upper()
    objindex = data.dtls.index(objkey)
    objcomm = commdct[objindex]
    adistutypefield = "Air Terminal Object Type"
    ifield = getfieldindex(data, commdct, objkey, adistutypefield)
    adistus = objcomm[ifield]['key']
    return adistus

def makeadistu_inlets(data, commdct):
    """make the dict adistu_inlets"""
    adistus = getadistus(data, commdct)
    # assume that the inlet node has the words "Air Inlet Node Name"
    airinletnode = "Air Inlet Node Name"
    adistu_inlets = {}
    for adistu in adistus:
        objkey = adistu.upper()
        objindex = data.dtls.index(objkey)
        objcomm = commdct[objindex]
        airinlets = []
        for i, comm in enumerate(objcomm):
            try:
                if comm['field'][0].find(airinletnode) != -1:
                    airinlets.append(comm['field'][0])
            except KeyError, e:
                pass
        adistu_inlets[adistu] = airinlets
    return adistu_inlets



"""routines to deal with tables in the html output files of E+
Should contain only routines that deal with html"""

# Copyright (C) 2009 Santosh Philip
# This file is part of tablediff.
# 
# tablediff is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# tablediff is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with tablediff.  If not, see <http://www.gnu.org/licenses/>.

from bs4 import BeautifulSoup, NavigableString, Tag

def nestedtable(table):
    """test if the table has nested tables"""
    if table.table.table == None:
        return False
    else:
        return True

class NotSimpleTable(Exception):
    pass
        

def is_simpletable(table):
    """test if the table has only strings in the cells"""
    tds = table('td')
    for td in tds:
        if td.contents != []:
            if len(td.contents) == 1:
                if not isinstance(td.contents[0], NavigableString):
                    return False
            else:
                return False
    return True

def table2matrix(table):
    """convert a table to a list of lists - a 2D matrix"""
    if not is_simpletable(table):
        raise NotSimpleTable, "Not able read a cell in the table as a string"
    rows = []
    for tr in table('tr'):
        row = []
        for td in tr('td'):
            try:
                row.append(td.contents[0])
            except IndexError, e:
                row.append('')
        rows.append(row)
    return rows

#=============================
# matrix operations

def equal_matrixsize(mat1, mat2):
    """return true if the mat1 and mat2 have same configuration of cells"""
    if mat1 == mat2: return True 
    if len(mat1) == len(mat2):
        for i in range(len(mat1)):
            if len(mat1[i]) == len(mat2[i]):
                pass
            else:
                return False
    else:
        return False
    return True


def gettables(soupbody):
    """from soupbody get all the tables"""
    body = soupbody
    tables = []        
    for obj in body.contents:
        try:
            if obj.name == 'table':
                tables.append(obj)
        except AttributeError, e:
            pass
    return tables
    

def flattenkey(btabledct):
    """flattens the key by bringing the header one level down into next dct"""
    decdct= {}
    for boldkey in btabledct.keys():
        for gridkey in btabledct[boldkey].keys():
            lst = [boldkey] + list(gridkey)
            newkey = tuple(lst)
            decdct[newkey] = btabledct[boldkey][gridkey]
    return decdct    

def bigsmallequal((value, isdiff)):
    """test if the value is big, small or equal, or None"""
    if isdiff == True:
        return 'big'
    if isdiff == False:
        try:
            anum = float(value)
            if anum != 0:
                return 'small'
            else:
                return 'equal'
        except ValueError, e:
            return None

def matrix2table3(table, mat, theformat='%s'):
    # TODO : unit test needed for this function
    """pushes the matrix into the table
    The matrix and table dimensions have to match. No checks are done
    returns a dictionary: {'big':n1, 'small':n2, 'equal':n3, None:n4} n1,n2.. are the sums
    The table is changed in place"""
    # TODO - test that table and matrix dimensions match
    # TODO - test that there are no nested tables
    thesum = {'big':0, 'small':0, 'equal':0, None:0, 'table_size_error':0}
    flat = []
    for row in mat:
        flat = flat + row
    alltds = table.findAll('td')
    for old, new in zip(alltds, flat):
        valuetype = bigsmallequal(new)
        thesum[valuetype] += 1 
        if valuetype:
            old['class'] = valuetype
        try:
            if isinstance(new[0], tuple):
                newval = '%s vs %s' % (new[0][0], new[0][1])
            else:
                newval = new[0]
                if valuetype != None:
                    newval = theformat % (newval,)                    
            old.contents[0].replaceWith(unicode(newval))
        except IndexError, e: # no contents to replace
            pass 
    return thesum           

def error2table(table):
    """pushes the table_size_error into the table
    The matrix and table dimensions have to match. No checks are done
    returns a dictionary: {'big':0, 'small':0, 'equal':0, None:0, 'table_size_error':1}
    The table is changed in place"""
    valuetype = 'table_size_error'
    thesum = {'big':0, 'small':0, 'equal':0, None:0, 'table_size_error':1}
    alltds = table.findAll('td')
    for td in alltds:
        td['class'] = valuetype
        try:
            td.contents[0].replaceWith(unicode(valuetype))
        except IndexError, e: # no contents to replace
            pass 
    # put in error message
    parent = table.parent
    index = parent.contents.index(table)
    themessage = 'table_size_error = The size of the tables in the files do not match'
    parent.insert(index + 1, '<em>%s</em>' % (themessage, ))
    
    return thesum           

    


# Copyright (c) 2012 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

"""make plant loop snippets"""
import sys
import copy
import bunch_subclass
import modeleditor
from modeleditor import IDF


class SomeFields(object):
    c_fields = ['Condenser Side Inlet Node Name',
    'Condenser Side Outlet Node Name',
    'Condenser Side Branch List Name',
    'Condenser Side Connector List Name',
    'Demand Side Inlet Node Name',
    'Demand Side Outlet Node Name',
    'Condenser Demand Side Branch List Name',
    'Condenser Demand Side Connector List Name']
    p_fields = ['Plant Side Inlet Node Name',
    'Plant Side Outlet Node Name',
    'Plant Side Branch List Name',
    'Plant Side Connector List Name',
    'Demand Side Inlet Node Name',
    'Demand Side Outlet Node Name',
    'Demand Side Branch List Name',
    'Demand Side Connector List Name']
    a_fields = ['Branch List Name',
    'Connector List Name',
    'Supply Side Inlet Node Name',
    'Demand Side Outlet Node Name',
    'Demand Side Inlet Node Names',
    'Supply Side Outlet Node Names']

    
        
def flattencopy(lst):
    """flatten and return a copy of the list
    indefficient on large lists"""
    # modified from http://stackoverflow.com/questions/2158395/flatten-an-irregular-list-of-lists-in-python
    TheList = copy.deepcopy(lst)
    listIsNested = True
    while listIsNested:                 #outer loop
        keepChecking = False
        Temp = []
        for element in TheList:         #inner loop
            if isinstance(element,list):
                Temp.extend(element)
                keepChecking = True
            else:
                Temp.append(element)
        listIsNested = keepChecking     #determine if outer loop exits
        TheList = Temp[:]
    return TheList

def test_flattencopy():
    """py.test for flattencopy"""
    tdata = (([1,2], [1,2]), #lst , nlst
    ([1,2,[3,4]], [1,2,3,4]), #lst , nlst
    ([1,2,[3,[4,5,6],7,8]], [1,2,3,4,5,6,7,8]), #lst , nlst
    ([1,2,[3,[4,5,[6,7],8],9]], [1,2,3,4,5,6,7,8,9]), #lst , nlst
    )
    for lst , nlst in tdata:
        result = flattencopy(lst)
        assert result == nlst
        
def makepipecomponent(idf, pname):
    """make a pipe component
    generate inlet outlet names"""
    apipe = idf.newidfobject("Pipe:Adiabatic".upper(), pname)
    apipe.Inlet_Node_Name = "%s_inlet" % (pname, )
    apipe.Outlet_Node_Name = "%s_outlet" % (pname, )
    return apipe
    
def makeductcomponent(idf, dname):
    """make a duct component
    generate inlet outlet names"""
    aduct = idf.newidfobject("duct".upper(), dname)
    aduct.Inlet_Node_Name = "%s_inlet" % (dname, )
    aduct.Outlet_Node_Name = "%s_outlet" % (dname, )
    return aduct
    
def makepipebranch(idf ,bname):
    """make a branch with a pipe
    use standard inlet outlet names"""
    # make the pipe component first
    pname = "%s_pipe" % (bname, )
    apipe = makepipecomponent(idf, pname)
    # now make the branch with the pipe in it
    abranch = idf.newidfobject("BRANCH", bname)    
    abranch.Component_1_Object_Type = 'Pipe:Adiabatic'
    abranch.Component_1_Name = pname
    abranch.Component_1_Inlet_Node_Name = apipe.Inlet_Node_Name
    abranch.Component_1_Outlet_Node_Name = apipe.Outlet_Node_Name
    abranch.Component_1_Branch_Control_Type = "Bypass"
    return abranch
    
def makeductbranch(idf ,bname):
    """make a branch with a duct
    use standard inlet outlet names"""
    # make the duct component first
    pname = "%s_duct" % (bname, )
    aduct = makeductcomponent(idf, pname)
    # now make the branch with the duct in it
    abranch = idf.newidfobject("BRANCH", bname)    
    abranch.Component_1_Object_Type = 'duct'
    abranch.Component_1_Name = pname
    abranch.Component_1_Inlet_Node_Name = aduct.Inlet_Node_Name
    abranch.Component_1_Outlet_Node_Name = aduct.Outlet_Node_Name
    abranch.Component_1_Branch_Control_Type = "Bypass"
    return abranch
    
def makeplantloop(idf, loopname, sloop, dloop):
    """make plant loop with pip components"""

    newplantloop = idf.newidfobject("PLANTLOOP", loopname)

    fields = SomeFields.p_fields

    # for use in bunch
    flnames = [field.replace(' ', '_') for field in fields]

    # implify naming
    fields1 = [field.replace('Plant Side', 'Supply') for field in fields]
    fields1 = [field.replace('Demand Side', 'Demand') for field in fields1]
    fields1 = [field[:field.find('Name') - 1] for field in fields1]
    fields1 = [field.replace(' Node', '') for field in fields1]
    fields1 = [field.replace(' List', 's') for field in fields1]
    # changesnames to 
    # ['Supply Inlet',
    #  'Supply Outlet',
    #  'Supply Branchs',
    #  'Supply Connectors',
    #  'Demand Inlet',
    #  'Demand Outlet',
    #  'Demand Branchs',
    #  'Demand Connectors']

    # TODO : pop connectors if no parallel branches
    # make fieldnames in the plant loop
    fieldnames = ['%s %s' % (loopname, field) for field in fields1]
    for fieldname, thefield in zip(fieldnames, flnames):
        newplantloop[thefield] = fieldname
    
    # make the branch lists for this plant loop    
    sbranchlist = idf.newidfobject("BRANCHLIST",
                    newplantloop.Plant_Side_Branch_List_Name)
    dbranchlist = idf.newidfobject("BRANCHLIST",
                    newplantloop.Demand_Side_Branch_List_Name)

    # add branch names to the branchlist
    sbranchnames = flattencopy(sloop)
    # sbranchnames = sloop[1]
    for branchname in sloop[1]:
        sbranchlist.obj.append(branchname)
    dbranchnames = flattencopy(dloop)
    # dbranchnames = dloop[1]
    for branchname in dloop[1]:
        dbranchlist.obj.append(branchname)

    # make a pipe branch for all branches in the loop

    # supply side
    sbranchs = []
    for bname in sbranchnames:
        branch = makepipebranch(idf, bname)
        sbranchs.append(branch)
    # rename inlet outlet of endpoints of loop
    anode = "Component_1_Inlet_Node_Name"
    sameinnode = "Plant_Side_Inlet_Node_Name"
    sbranchs[0][anode] =  newplantloop[sameinnode]
    anode = "Component_1_Outlet_Node_Name"
    sameoutnode = "Plant_Side_Outlet_Node_Name"
    sbranchs[-1][anode] =  newplantloop[sameoutnode]
    # rename inlet outlet of endpoints of loop - rename in pipe
    pname = sbranchs[0]['Component_1_Name'] # get the pipe name
    apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    apipe.Inlet_Node_Name = newplantloop[sameinnode]
    pname = sbranchs[-1]['Component_1_Name'] # get the pipe name
    apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    apipe.Outlet_Node_Name = newplantloop[sameoutnode]

    # demand side
    dbranchs = []
    for bname in dbranchnames:
        branch = makepipebranch(idf, bname)
        dbranchs.append(branch)
    # rename inlet outlet of endpoints of loop - rename in branch
    anode = "Component_1_Inlet_Node_Name"
    sameinnode = "Demand_Side_Inlet_Node_Name"
    dbranchs[0][anode] =  newplantloop[sameinnode]
    anode = "Component_1_Outlet_Node_Name"
    sameoutnode = "Demand_Side_Outlet_Node_Name"
    dbranchs[-1][anode] =  newplantloop[sameoutnode]
    # rename inlet outlet of endpoints of loop - rename in pipe
    pname = dbranchs[0]['Component_1_Name'] # get the pipe name
    apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    apipe.Inlet_Node_Name = newplantloop[sameinnode]
    pname = dbranchs[-1]['Component_1_Name'] # get the pipe name
    apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    apipe.Outlet_Node_Name = newplantloop[sameoutnode]


    # TODO : test if there are parallel branches
    # make the connectorlist an fill fields
    sconnlist = idf.newidfobject("CONNECTORLIST",
                    newplantloop.Plant_Side_Connector_List_Name)
    sconnlist.Connector_1_Object_Type = "Connector:Splitter"
    sconnlist.Connector_1_Name = "%s_supply_splitter" % (loopname, )
    sconnlist.Connector_2_Object_Type = "Connector:Mixer"
    sconnlist.Connector_2_Name = "%s_supply_mixer" % (loopname, )
    dconnlist = idf.newidfobject("CONNECTORLIST",
                    newplantloop.Demand_Side_Connector_List_Name)
    dconnlist.Connector_1_Object_Type = "Connector:Splitter"
    dconnlist.Connector_1_Name = "%s_demand_splitter" % (loopname, )
    dconnlist.Connector_2_Object_Type = "Connector:Mixer"
    dconnlist.Connector_2_Name = "%s_demand_mixer" % (loopname, )

    # make splitters and mixers
    s_splitter = idf.newidfobject("CONNECTOR:SPLITTER", 
        sconnlist.Connector_1_Name)
    s_splitter.obj.extend([sloop[0]] + sloop[1])
    s_mixer = idf.newidfobject("CONNECTOR:MIXER", 
        sconnlist.Connector_2_Name)
    s_mixer.obj.extend([sloop[-1]] + sloop[1])
    # -
    d_splitter = idf.newidfobject("CONNECTOR:SPLITTER", 
        dconnlist.Connector_1_Name)
    d_splitter.obj.extend([dloop[0]] + dloop[1])
    d_mixer = idf.newidfobject("CONNECTOR:MIXER", 
        dconnlist.Connector_2_Name)
    d_mixer.obj.extend([dloop[-1]] + dloop[1])
    return newplantloop

def getbranchcomponents(idf, branch, utest=False):
    """get the components of the branch"""
    fobjtype = 'Component_%s_Object_Type'
    fobjname = 'Component_%s_Name'
    complist = []
    for i in xrange(1, 100000):
        try:
            objtype = branch[fobjtype % (i, )]
            if objtype.strip() == '':
                break
            objname = branch[fobjname % (i, )]
            complist.append((objtype, objname))
        except bunch_subclass.BadEPFieldError, e:
            break
    if utest:
        return complist
    else:
        return [idf.getobject(ot, on) for ot, on in complist]

def renamenodes(idf, fieldtype):
    """rename all the changed nodes"""
    renameds = []
    for key in idf.model.dtls:
        for idfobject in idf.idfobjects[key]:
            for fieldvalue in idfobject.obj:
                if type(fieldvalue) is list:
                    if fieldvalue not in renameds:
                        cpvalue = copy.copy(fieldvalue)
                        renameds.append(cpvalue)

    # do the renaming
    for key in idf.model.dtls:
        for idfobject in idf.idfobjects[key]:
            for i, fieldvalue in enumerate(idfobject.obj):
                itsidd = idfobject.objidd[i]
                if itsidd.has_key('type'):
                    if itsidd['type'][0] == fieldtype:
                        tempdct = dict(renameds)
                        if type(fieldvalue) is list:
                            fieldvalue = fieldvalue[-1]
                            idfobject.obj[i] = fieldvalue
                        else:
                            if tempdct.has_key(fieldvalue):
                                fieldvalue = tempdct[fieldvalue]
                                idfobject.obj[i] = fieldvalue
    
def getfieldnamesendswith(idfobject, endswith):
    """get the filednames for the idfobject based on endswith"""
    objls = idfobject.objls
    return [name for name in objls if name.endswith(endswith)]
    
def getnodefieldname(idfobject, endswith, fluid=''):
    """return the field name of the node
    fluid is only needed if there are air and water nodes
    fluid is Air or Water or ''. 
    if the fluid is Steam, use Water"""
    nodenames = getfieldnamesendswith(idfobject, endswith)
    fnodenames=[nd for nd in nodenames if nd.find(fluid)!=-1]
    if len(fnodenames) == 0:
        nodename = nodenames[0]
    else:
        nodename = fnodenames[0]
    return nodename
        
    
def connectcomponents(idf, components, fluid=''):
    """rename nodes so that the components get connected
    fluid is only needed if there are air and water nodes
    fluid is Air or Water or ''. 
    if the fluid is Steam, use Water"""
    for i in range(len(components) - 1):
        thiscomp = components[i]
        nextcomp = components[i + 1]
        initinletoutlet(idf, thiscomp, force=False)
        initinletoutlet(idf, nextcomp, force=False)
        betweennodename = "%s_%s_node" % (thiscomp.Name, nextcomp.Name)
        outletnodename = getnodefieldname(thiscomp, "Outlet_Node_Name", fluid)
        thiscomp[outletnodename] = [thiscomp[outletnodename], betweennodename]
        inletnodename = getnodefieldname(nextcomp, "Inlet_Node_Name", fluid)
        nextcomp[inletnodename] = [nextcomp[inletnodename], betweennodename]
    return components
     
    
def initinletoutlet(idf, idfobject, force=False):
    """initialze values for all the inlet outlet nodes for the object.
    if force == False, it willl init only if field = '' """
    def blankfield(fieldvalue):
        """test for blank field"""
        try:
            if fieldvalue.strip() == '':
                return True
            else:
                return False
        except AttributeError, e: # field may be a list
            return False
            
    inletfields = getfieldnamesendswith(idfobject, "Inlet_Node_Name")
    for inletfield in inletfields:
        if blankfield(idfobject[inletfield]) == True or force == True:
            idfobject[inletfield] = "%s_%s" % (idfobject.Name, inletfield)
    outletfields = getfieldnamesendswith(idfobject, "Outlet_Node_Name")
    for outletfield in outletfields:
        if blankfield(idfobject[outletfield]) == True or force == True:
            idfobject[outletfield] = "%s_%s" % (idfobject.Name, outletfield)
    return idfobject
    
def componentsintobranch(idf, branch, componentlist, fluid=''):
    """insert a list of components into a branch
    fluid is only needed if there are air and water nodes in same object
    fluid is Air or Water or ''. 
    if the fluid is Steam, use Water"""
    # assumes that the nodes of the component connect to each other
    # empty branch if it has existing components
    thebranchname = branch.Name
    thebranch = idf.removeextensibles('BRANCH', thebranchname) # empty the branch
    # fill in the new components with the node names into this branch
        # find the first extensible field and fill in the data in obj.
    e_index = idf.getextensibleindex('BRANCH', thebranchname)
    theobj = thebranch.obj
    modeleditor.extendlist(theobj, e_index) # just being careful here
    for comp in componentlist:
        theobj.append(comp.key)
        theobj.append(comp.Name)
        inletnodename = getnodefieldname(comp, "Inlet_Node_Name", fluid)
        theobj.append(comp[inletnodename])
        outletnodename = getnodefieldname(comp, "Outlet_Node_Name", fluid)
        theobj.append(comp[outletnodename])
        theobj.append('')
    
    return thebranch

def replacebranch(idf, loop, branch, 
                listofcomponents, fluid='', debugsave=False):
    """It will replace the components in the branch with components in 
    listofcomponents"""
    # join them into a branch
    # -----------------------
    # np1_inlet -> np1 -> np1_np2_node -> np2 -> np2_outlet
        # change the node names in the component
        # empty the old branch
        # fill in the new components with the node names into this branch

    components = listofcomponents
    connectcomponents(idf, components, fluid=fluid)
    if debugsave:
        idf.saveas("hhh3.idf")

    thebranch = branch
    componentsintobranch(idf, thebranch, components, fluid=fluid)

    # # gather all renamed nodes
    # # do the renaming
    renamenodes(idf, 'node')
    if debugsave:
        idf.saveas("hhh7.idf")

    # check for the end nodes of the loop
    if loop.key == 'PLANTLOOP':
        fields = SomeFields.p_fields
    if loop.key == 'CONDENSERLOOP':
        fields = SomeFields.c_fields
    # for use in bunch
    flnames = [field.replace(' ', '_') for field in fields]
    
    supplyconlistname = loop[flnames[3]] # Plant_Side_Connector_List_Name
    supplyconlist = idf.getobject('CONNECTORLIST', supplyconlistname)
    for i in xrange(1, 100000): # large range to hit end
        try:
            fieldname = 'Connector_%s_Object_Type' % (i, )
            ctype = supplyconlist[fieldname]
        except bunch_subclass.BadEPFieldError, e:
            break
        if ctype.strip() == '':
            break
        fieldname = 'Connector_%s_Name' % (i, )
        cname = supplyconlist[fieldname]
        connector = idf.getobject(ctype.upper(), cname)
        if connector.key == 'CONNECTOR:SPLITTER':
            firstbranchname = connector.Inlet_Branch_Name
            cbranchname = firstbranchname
            isfirst = True
        if connector.key == 'CONNECTOR:MIXER':
            lastbranchname = connector.Outlet_Branch_Name
            cbranchname = lastbranchname
            isfirst = False
        if cbranchname == thebranch.Name:
            # rename end nodes
            comps = getbranchcomponents(idf, thebranch)
            if isfirst:
                comp = comps[0]
                inletnodename = getnodefieldname(comp, 
                                        "Inlet_Node_Name", fluid)
                comp[inletnodename] = [comp[inletnodename],
                                loop[flnames[0]]] # Plant_Side_Inlet_Node_Name
            else:
                comp = comps[-1]
                outletnodename = getnodefieldname(comp, 
                                        "Outlet_Node_Name", fluid)
                comp[outletnodename] = [comp[outletnodename], 
                            loop[flnames[1]]] # .Plant_Side_Outlet_Node_Name

    demandconlistname = loop[flnames[7]] # .Demand_Side_Connector_List_Name
    demandconlist = idf.getobject('CONNECTORLIST', demandconlistname)
    for i in xrange(1, 100000): # large range to hit end
        try:
            fieldname = 'Connector_%s_Object_Type' % (i, )
            ctype = demandconlist[fieldname]
        except bunch_subclass.BadEPFieldError, e:
            break
        if ctype.strip() == '':
            break
        fieldname = 'Connector_%s_Name' % (i, )
        cname = demandconlist[fieldname]
        connector = idf.getobject(ctype.upper(), cname)
        if connector.key == 'CONNECTOR:SPLITTER':
            firstbranchname = connector.Inlet_Branch_Name
            cbranchname = firstbranchname
            isfirst = True
        if connector.key == 'CONNECTOR:MIXER':
            lastbranchname = connector.Outlet_Branch_Name 
            cbranchname = lastbranchname
            isfirst = False
        # print cbranchname, thebranch
        if cbranchname == thebranch.Name:
            # rename end nodes
            comps = getbranchcomponents(idf, thebranch)
            if isfirst:
                comp = comps[0]
                inletnodename = getnodefieldname(comp, 
                                        "Inlet_Node_Name", fluid)
                comp[inletnodename] = [comp[inletnodename],
                            loop[flnames[4]]] #.Demand_Side_Inlet_Node_Name
            if not isfirst:
                comp = comps[-1]
                outletnodename = getnodefieldname(comp, 
                                        "Outlet_Node_Name", fluid)
                comp[outletnodename] = [comp[outletnodename], 
                            loop[flnames[5]]] # .Demand_Side_Outlet_Node_Name


    if debugsave:
        idf.saveas("hhh8.idf")

    # # gather all renamed nodes
    # # do the renaming
    renamenodes(idf, 'node')
    if debugsave:
        idf.saveas("hhh9.idf")
    return thebranch



def makecondenserloop(idf, loopname, sloop, dloop):
    """make condenser loop with pipe components"""

    newcondenserloop = idf.newidfobject("CondenserLoop".upper(), loopname)

    fields = SomeFields.c_fields

    # for use in bunch
    flnames = [field.replace(' ', '_') for field in fields]

    # TODO : check if these change too.
    # simplify naming
    fields1 = [field.replace('Condenser Side', 
                                    'Cond_Supply') for field in fields]
    fields1 = [field.replace('Demand Side', 'Demand') for field in fields1]
    fields1 = [field[:field.find('Name') - 1] for field in fields1]
    fields1 = [field.replace(' Node', '') for field in fields1]
    fields1 = [field.replace(' List', 's') for field in fields1]
    # changesnames to 
    # ['Cond_Supply Inlet',
    #  'Cond_Supply Outlet',
    #  'Cond_Supply Branchs',
    #  'Cond_Supply Connectors',
    #  'Demand Inlet',
    #  'Demand Outlet',
    #  'Demand Branchs',
    #  'Demand Connectors']

    # old TODO : pop connectors if no parallel branches
    # make fieldnames in the condenser loop
    fieldnames = ['%s %s' % (loopname, field) for field in fields1]
    for fieldname, thefield in zip(fieldnames, flnames):
        newcondenserloop[thefield] = fieldname
    
    # make the branch lists for this condenser loop    
    sbranchlist = idf.newidfobject("BRANCHLIST",
                    newcondenserloop.Condenser_Side_Branch_List_Name)
    dbranchlist = idf.newidfobject("BRANCHLIST",
                    newcondenserloop.Condenser_Demand_Side_Branch_List_Name)

    # add branch names to the branchlist
    sbranchnames = flattencopy(sloop)
    # sbranchnames = sloop[1]
    for branchname in sloop[1]:
        sbranchlist.obj.append(branchname)
    dbranchnames = flattencopy(dloop)
    # dbranchnames = dloop[1]
    for branchname in dloop[1]:
        dbranchlist.obj.append(branchname)

    # make a pipe branch for all branches in the loop

    # supply side
    sbranchs = []
    for bname in sbranchnames:
        branch = makepipebranch(idf, bname)
        sbranchs.append(branch)
    # rename inlet outlet of endpoints of loop
    anode = "Component_1_Inlet_Node_Name"
    sameinnode = "Condenser_Side_Inlet_Node_Name" # TODO : change ?
    sbranchs[0][anode] =  newcondenserloop[sameinnode]
    anode = "Component_1_Outlet_Node_Name"
    sameoutnode = "Condenser_Side_Outlet_Node_Name" # TODO : change ?
    sbranchs[-1][anode] =  newcondenserloop[sameoutnode]
    # rename inlet outlet of endpoints of loop - rename in pipe
    pname = sbranchs[0]['Component_1_Name'] # get the pipe name
    apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    apipe.Inlet_Node_Name = newcondenserloop[sameinnode]
    pname = sbranchs[-1]['Component_1_Name'] # get the pipe name
    apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    apipe.Outlet_Node_Name = newcondenserloop[sameoutnode]

    # demand side
    dbranchs = []
    for bname in dbranchnames:
        branch = makepipebranch(idf, bname)
        dbranchs.append(branch)
    # rename inlet outlet of endpoints of loop - rename in branch
    anode = "Component_1_Inlet_Node_Name"
    sameinnode = "Demand_Side_Inlet_Node_Name" # TODO : change ?
    dbranchs[0][anode] =  newcondenserloop[sameinnode]
    anode = "Component_1_Outlet_Node_Name"
    sameoutnode = "Demand_Side_Outlet_Node_Name" # TODO : change ?
    dbranchs[-1][anode] =  newcondenserloop[sameoutnode]
    # rename inlet outlet of endpoints of loop - rename in pipe
    pname = dbranchs[0]['Component_1_Name'] # get the pipe name
    apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    apipe.Inlet_Node_Name = newcondenserloop[sameinnode]
    pname = dbranchs[-1]['Component_1_Name'] # get the pipe name
    apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    apipe.Outlet_Node_Name = newcondenserloop[sameoutnode]


    # TODO : test if there are parallel branches
    # make the connectorlist an fill fields
    sconnlist = idf.newidfobject("CONNECTORLIST",
                    newcondenserloop.Condenser_Side_Connector_List_Name)
    sconnlist.Connector_1_Object_Type = "Connector:Splitter"
    sconnlist.Connector_1_Name = "%s_supply_splitter" % (loopname, )
    sconnlist.Connector_2_Object_Type = "Connector:Mixer"
    sconnlist.Connector_2_Name = "%s_supply_mixer" % (loopname, )
    dconnlist = idf.newidfobject("CONNECTORLIST",
        newcondenserloop.Condenser_Demand_Side_Connector_List_Name)
    dconnlist.Connector_1_Object_Type = "Connector:Splitter"
    dconnlist.Connector_1_Name = "%s_demand_splitter" % (loopname, )
    dconnlist.Connector_2_Object_Type = "Connector:Mixer"
    dconnlist.Connector_2_Name = "%s_demand_mixer" % (loopname, )

    # make splitters and mixers
    s_splitter = idf.newidfobject("CONNECTOR:SPLITTER", 
        sconnlist.Connector_1_Name)
    s_splitter.obj.extend([sloop[0]] + sloop[1])
    s_mixer = idf.newidfobject("CONNECTOR:MIXER", 
        sconnlist.Connector_2_Name)
    s_mixer.obj.extend([sloop[-1]] + sloop[1])
    # -
    d_splitter = idf.newidfobject("CONNECTOR:SPLITTER", 
        dconnlist.Connector_1_Name)
    d_splitter.obj.extend([dloop[0]] + dloop[1])
    d_mixer = idf.newidfobject("CONNECTOR:MIXER", 
        dconnlist.Connector_2_Name)
    d_mixer.obj.extend([dloop[-1]] + dloop[1])
    return newcondenserloop

def makeairloop(idf, loopname, sloop, dloop):
    """make an airloop"""
    newairloop = idf.newidfobject("AirLoopHVAC".upper(), loopname)



    fields = SomeFields.a_fields

    # for use in bunch
    flnames = [field.replace(' ', '_') for field in fields]

    # TODO : check if these change too.
    # simplify naming
    a_fields = ['Branch List Name',
     'Connector List Name',
     'Supply Side Inlet Node Name',
     'Demand Side Outlet Node Name',
     'Demand Side Inlet Node Names',
     'Supply Side Outlet Node Names']
    # changesnames to
    # from
    # ['Branch List Name',
    #  'Connector List Name',
    #  'Supply Side Inlet Node Name',
    #  'Demand Side Outlet Node Name',
    #  'Demand Side Inlet Node Names',
    #  'Supply Side Outlet Node Names']
    # to (airloop=a_loop)
    # ['Branches',
    #  'Connectors',
    #  'Supply Inlet',
    #  'Demand Outlet',
    #  'Demand Inlet',
    #  'Supply Outlet'] 
    fields1 = ['Branches',
     'Connectors',
     'Supply Inlet',
     'Demand Outlet',
     'Demand Inlet',
     'Supply Outlet'] 

    # old TODO : pop connectors if no parallel branches
    # make fieldnames in the air loop
    fieldnames = ['%s %s' % (loopname, field) for field in fields1]
    for fieldname, thefield in zip(fieldnames, flnames):
        newairloop[thefield] = fieldname

    # make the branch lists for this air loop    
    sbranchlist = idf.newidfobject("BRANCHLIST",
                    newairloop[flnames[0]])

    # not needed in airloop                
    # dbranchlist = idf.newidfobject("BRANCHLIST",
    #                 newcondenserloop.Condenser_Demand_Side_Branch_List_Name)

    # add branch names to the branchlist
    sbranchnames = flattencopy(sloop)
    # sbranchnames = sloop[1]
    for branchname in sbranchnames:
        sbranchlist.obj.append(branchname)
    # dbranchnames = flattencopy(dloop)
    # # dbranchnames = dloop[1]
    # for branchname in dloop[1]:
    #     dbranchlist.obj.append(branchname)
    # 
    # # make a duct branch for all branches in the loop
    # 
    # supply side
    sbranchs = []
    for bname in sbranchnames:
        branch = makeductbranch(idf, bname)
        sbranchs.append(branch)
    # rename inlet outlet of endpoints of loop
    anode = "Component_1_Inlet_Node_Name"
    sameinnode = "Supply_Side_Inlet_Node_Name" # TODO : change ?
    sbranchs[0][anode] =  newairloop[sameinnode]
    anode = "Component_1_Outlet_Node_Name"
    sameoutnode = "Supply_Side_Outlet_Node_Names" # TODO : change ?
    sbranchs[-1][anode] =  newairloop[sameoutnode]
    # rename inlet outlet of endpoints of loop - rename in pipe
    dname = sbranchs[0]['Component_1_Name'] # get the duct name
    aduct = idf.getobject('duct'.upper(), dname) # get duct
    aduct.Inlet_Node_Name = newairloop[sameinnode]
    dname = sbranchs[-1]['Component_1_Name'] # get the duct name
    aduct = idf.getobject('duct'.upper(), dname) # get duct
    aduct.Outlet_Node_Name = newairloop[sameoutnode]
    # 
    # # demand side
    # dbranchs = []
    # for bname in dbranchnames:
    #     branch = makepipebranch(idf, bname)
    #     dbranchs.append(branch)
    # # rename inlet outlet of endpoints of loop - rename in branch
    # anode = "Component_1_Inlet_Node_Name"
    # sameinnode = "Demand_Side_Inlet_Node_Name" # TODO : change ?
    # dbranchs[0][anode] =  newcondenserloop[sameinnode]
    # anode = "Component_1_Outlet_Node_Name"
    # sameoutnode = "Demand_Side_Outlet_Node_Name" # TODO : change ?
    # dbranchs[-1][anode] =  newcondenserloop[sameoutnode]
    # # rename inlet outlet of endpoints of loop - rename in pipe
    # pname = dbranchs[0]['Component_1_Name'] # get the pipe name
    # apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    # apipe.Inlet_Node_Name = newcondenserloop[sameinnode]
    # pname = dbranchs[-1]['Component_1_Name'] # get the pipe name
    # apipe = idf.getobject('Pipe:Adiabatic'.upper(), pname) # get pipe
    # apipe.Outlet_Node_Name = newcondenserloop[sameoutnode]
    # 
    # 
    # # TODO : test if there are parallel branches
    # make the connectorlist an fill fields
    sconnlist = idf.newidfobject("CONNECTORLIST",
                    newairloop.Connector_List_Name)
    sconnlist.Connector_1_Object_Type = "Connector:Splitter"
    sconnlist.Connector_1_Name = "%s_supply_splitter" % (loopname, )
    sconnlist.Connector_2_Object_Type = "Connector:Mixer"
    sconnlist.Connector_2_Name = "%s_supply_mixer" % (loopname, )
    # dconnlist = idf.newidfobject("CONNECTORLIST",
    #     newcondenserloop.Condenser_Demand_Side_Connector_List_Name)
    # dconnlist.Connector_1_Object_Type = "Connector:Splitter"
    # dconnlist.Connector_1_Name = "%s_demand_splitter" % (loopname, )
    # dconnlist.Connector_2_Object_Type = "Connector:Mixer"
    # dconnlist.Connector_2_Name = "%s_demand_mixer" % (loopname, )
    # 
    # make splitters and mixers
    s_splitter = idf.newidfobject("CONNECTOR:SPLITTER", 
        sconnlist.Connector_1_Name)
    s_splitter.obj.extend([sloop[0]] + sloop[1])
    s_mixer = idf.newidfobject("CONNECTOR:MIXER", 
        sconnlist.Connector_2_Name)
    s_mixer.obj.extend([sloop[-1]] + sloop[1])
    # # -
    # d_splitter = idf.newidfobject("CONNECTOR:SPLITTER", 
    #     dconnlist.Connector_1_Name)
    # d_splitter.obj.extend([dloop[0]] + dloop[1])
    # d_mixer = idf.newidfobject("CONNECTOR:MIXER", 
    #     dconnlist.Connector_2_Name)
    # d_mixer.obj.extend([dloop[-1]] + dloop[1])
    # return newcondenserloop

    # demand side loop for airloop is made below 
    #ZoneHVAC:EquipmentConnections
    for zone in dloop:
        equipconn = idf.newidfobject("ZoneHVAC:EquipmentConnections".upper())
        equipconn.Zone_Name = zone
        fldname = "Zone_Conditioning_Equipment_List_Name"
        equipconn[fldname] = "%s equip list" % (zone, )
        fldname = "Zone_Air_Inlet_Node_or_NodeList_Name"
        equipconn[fldname] = "%s Inlet Node" % (zone, )
        fldname = "Zone_Air_Node_Name"
        equipconn[fldname] = "%s Node" % (zone, )
        fldname = "Zone_Return_Air_Node_Name"
        equipconn[fldname] = "%s Outlet Node" % (zone, )

    # make ZoneHVAC:EquipmentList
    for zone in dloop:
        z_equiplst = idf.newidfobject("ZoneHVAC:EquipmentList".upper())
        z_equipconn = modeleditor.getobjects(idf.idfobjects, 
            idf.model, idf.idd_info, 
            "ZoneHVAC:EquipmentConnections".upper(), #places=7,
            **dict(Zone_Name=zone))[0]
        z_equiplst.Name = z_equipconn.Zone_Conditioning_Equipment_List_Name
        fld = "Zone_Equipment_1_Object_Type"
        z_equiplst[fld] = "AirTerminal:SingleDuct:Uncontrolled"
        z_equiplst.Zone_Equipment_1_Name = "%sDirectAir" % (zone, )
        z_equiplst.Zone_Equipment_1_Cooling_Sequence = 1
        z_equiplst.Zone_Equipment_1_Heating_or_NoLoad_Sequence = 1

    # make AirTerminal:SingleDuct:Uncontrolled
    for zone in dloop:
        z_equipconn = modeleditor.getobjects(idf.idfobjects, 
            idf.model, idf.idd_info, 
            "ZoneHVAC:EquipmentConnections".upper(), #places=7,
            **dict(Zone_Name=zone))[0]
        key = "AirTerminal:SingleDuct:Uncontrolled".upper()
        z_airterm = idf.newidfobject(key)
        z_airterm.Name = "%sDirectAir" % (zone, )
        fld1 = "Zone_Supply_Air_Node_Name"
        fld2 = "Zone_Air_Inlet_Node_or_NodeList_Name"
        z_airterm[fld1] = z_equipconn[fld2]
        z_airterm.Maximum_Air_Flow_Rate = 'autosize'

    # MAKE AirLoopHVAC:ZoneSplitter
    # zone = dloop[0]
    key = "AirLoopHVAC:ZoneSplitter".upper()
    z_splitter = idf.newidfobject(key)
    z_splitter.Name = "%s Demand Side Splitter" % (loopname, )
    z_splitter.Inlet_Node_Name = newairloop.Demand_Side_Inlet_Node_Names
    for i, zone in enumerate(dloop):
        z_equipconn = modeleditor.getobjects(idf.idfobjects, 
            idf.model, idf.idd_info, 
            "ZoneHVAC:EquipmentConnections".upper(), #places=7,
            **dict(Zone_Name=zone))[0]
        fld = "Outlet_%s_Node_Name" % (i + 1, )
        z_splitter[fld] = z_equipconn.Zone_Air_Inlet_Node_or_NodeList_Name

    # make AirLoopHVAC:SupplyPath
    key = "AirLoopHVAC:SupplyPath".upper()
    z_supplypth = idf.newidfobject(key)
    z_supplypth.Name = "%sSupplyPath" % (loopname, )
    fld1 = "Supply_Air_Path_Inlet_Node_Name"
    fld2 = "Demand_Side_Inlet_Node_Names"
    z_supplypth[fld1] = newairloop[fld2]
    z_supplypth.Component_1_Object_Type = "AirLoopHVAC:ZoneSplitter"
    z_supplypth.Component_1_Name = z_splitter.Name

    # make AirLoopHVAC:ZoneMixer
    key = "AirLoopHVAC:ZoneMixer".upper()
    z_mixer = idf.newidfobject(key)
    z_mixer.Name = "%s Demand Side Mixer" % (loopname, )
    z_mixer.Outlet_Node_Name = newairloop.Demand_Side_Outlet_Node_Name
    for i, zone in enumerate(dloop):
        z_equipconn = modeleditor.getobjects(idf.idfobjects, 
            idf.model, idf.idd_info, 
            "ZoneHVAC:EquipmentConnections".upper(), #places=7,
            **dict(Zone_Name=zone))[0]
        fld = "Inlet_%s_Node_Name" % (i + 1, )
        z_mixer[fld] = z_equipconn.Zone_Return_Air_Node_Name

    # make AirLoopHVAC:ReturnPath
    key = "AirLoopHVAC:ReturnPath".upper()
    z_returnpth = idf.newidfobject(key)
    z_returnpth.Name = "%sReturnPath" % (loopname, )
    z_returnpth.Return_Air_Path_Outlet_Node_Name = newairloop.Demand_Side_Outlet_Node_Name
    z_returnpth.Component_1_Object_Type = "AirLoopHVAC:ZoneMixer"
    z_returnpth.Component_1_Name = z_mixer.Name
    
def main():
    from StringIO import StringIO
    import iddv7
    IDF.setiddname(StringIO(iddv7.iddtxt))
    idf1 = IDF(StringIO(''))
    loopname = "p_loop"
    sloop = ['sb0', ['sb1', 'sb2', 'sb3'], 'sb4']
    dloop = ['db0', ['db1', 'db2', 'db3'], 'db4']
    # makeplantloop(idf1, loopname, sloop, dloop)
    loopname = "c_loop"
    sloop = ['sb0', ['sb1', 'sb2', 'sb3'], 'sb4']
    dloop = ['db0', ['db1', 'db2', 'db3'], 'db4']
    # makecondenserloop(idf1, loopname, sloop, dloop)
    loopname = "a_loop"
    sloop = ['sb0', ['sb1', 'sb2', 'sb3'], 'sb4']
    dloop = ['zone1', 'zone2', 'zone3']
    makeairloop(idf1, loopname, sloop, dloop)
    idf1.saveas("hh1.idf")


if __name__ == '__main__':
    main()



# Copyright (c) 2012 Tuan Tran

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

""""This module is used for calculation of zone volume for E+ surfaces"""
# Wrote by Tuan Tran trantuan@hawaii.edu / tranhuuanhtuan@gmail.com
# School of Architecture, University of Hawaii at Manoa

# Find the intersection between two lines
# V = (1/6)*|(a-d).((b-d)x(c-d))| http://en.wikipedia.org/wiki/Tetrahedron

import numpy as np

def vol_tehrahedron(poly):
    """volume of a irregular tetrahedron"""
    a = np.array(poly[0])
    b = np.array(poly[1])
    c = np.array(poly[2])
    d = np.array(poly[3])
    return abs(np.dot((a-d), np.cross((b-d),(c-d))) / 6)

def central_p(poly1,poly2):
    central_point = np.array([0.0, 0.0, 0.0])
    for i in range(len(poly1)):
        central_point += np.array(poly1[i]) + np.array(poly2[i])
    return central_point/ (len(poly1)) / 2

def vol(poly1,poly2):
    """"volume of a zone defined by two polygon bases """
    c_point = central_p(poly1, poly2)
    c_point = (c_point[0], c_point[1], c_point[2])
    vol_therah = 0
    N = len(poly1)
    poly1.append(poly1[0])
    poly2.append(poly2[0])
    for i in range(N-2):
        # the upper part
        tehrahedron = [c_point,poly1[0], poly1[i+1], poly1[i+2]]
        vol_therah += vol_tehrahedron(tehrahedron)
        # the bottom part
        tehrahedron = [c_point,poly2[0], poly2[i+1], poly2[i+2]]
        vol_therah += vol_tehrahedron(tehrahedron)
    # the middle part
    for i in range(N):
        tehrahedron = [c_point, poly1[i], poly2[i], poly2[i+1]]
        vol_therah += vol_tehrahedron(tehrahedron)
        tehrahedron = [c_point, poly1[i], poly1[i+1], poly2[i]]
        vol_therah += vol_tehrahedron(tehrahedron)
    return vol_therah


# Copyright (c) 2012 Tuan Tran

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

"""This module is used for assisted calculations on E+ surfaces"""
# Wrote by Tuan Tran trantuan@hawaii.edu / tranhuuanhtuan@gmail.com
# School of Architecture, University of Hawaii at Manoa

####### The following code within the block credited by ActiveState Code Recipes code.activestate.com
## {{{ http://code.activestate.com/recipes/578276/ (r1)

import numpy as np
import math

def area(poly):
    """Area of a polygon poly""" 
    if len(poly) < 3: # not a plane - no area
        return 0
    total = [0, 0, 0]
    N = len(poly)
    for i in range(N):
        vi1 = poly[i]
        vi2 = poly[(i+1) % N]
        prod = np.cross(vi1, vi2)
        total[0] += prod[0]
        total[1] += prod[1]
        total[2] += prod[2]
    result = np.dot(total, unit_normal(poly[0], poly[1], poly[2]))
    return abs(result/2)

#unit normal vector of plane defined by points a, b, and c
def unit_normal(a, b, c):
    x = np.linalg.det([[1,a[1],a[2]],[1,b[1],b[2]],[1,c[1],c[2]]])
    y = np.linalg.det([[a[0],1,a[2]],[b[0],1,b[2]],[c[0],1,c[2]]])
    z = np.linalg.det([[a[0],a[1],1],[b[0],b[1],1],[c[0],c[1],1]])
    magnitude = (x**2 + y**2 + z**2)**.5
    mag = (x/magnitude, y/magnitude, z/magnitude)
    if magnitude < 0.00000001: mag = (0,0,0)
    return mag
## end of http://code.activestate.com/recipes/578276/ }}}

# distance between two points
def dist(p1,p2):
    """Distance between two points"""
    return ((p2[0] - p1[0])**2 + (p2[1] - p1[1])**2 + (p2[2] - p1[2])**2)**0.5

# width of a rectangular polygon        
def width(poly):
    """Width of a polygon poly"""
    N = len(poly) - 1
    if abs(poly[N][2] - poly[0][2]) < abs(poly[1][2] - poly[0][2]):
        return dist(poly[N],poly[0])
    elif abs(poly[N][2] - poly[0][2]) > abs(poly[1][2] - poly[0][2]):
        return dist(poly[1], poly[0])
    else: return max(dist(poly[N], poly[0]), dist(poly[1], poly[0]))

# height of a polygon poly
def height(poly):
    """Height of a polygon poly"""
    N = len(poly) - 1
    if abs(poly[N][2] - poly[0][2]) > abs(poly[1][2] - poly[0][2]):
        return dist(poly[N],poly[0])
    elif abs(poly[N][2] - poly[0][2]) < abs(poly[1][2] - poly[0][2]):
        return dist(poly[1], poly[0])
    else: 
        return min(dist(poly[N], poly[0]), dist(poly[1], poly[0]))
    
# angle between two vectors
def angle2vecs(vec1,vec2):
    # vector a * vector b = |a|*|b|* cos(angle between vector a and vector b)
    dot = np.dot(vec1,vec2)
    vec1_modulus = np.sqrt((vec1*vec1).sum())
    vec2_modulus = np.sqrt((vec2*vec2).sum())
    if (vec1_modulus * vec2_modulus) == 0:
        cos_angle = 1
    else: cos_angle = dot / (vec1_modulus * vec2_modulus)
    return math.degrees(np.arccos(cos_angle)) 

# orienation of a polygon poly
def azimuth(poly):
    """Azimuth of a polygon poly"""
    N = len(poly) - 1
    vec = unit_normal(poly[0], poly[1], poly[N])
    vec_azi = np.array([vec[0], vec[1], 0])
    vec_N = np.array([0, 1, 0])
    # update by Santosh
    # angle2vecs gives the smallest angle between the vectors
    # so for a west wall angle2vecs will give 90
    # the following 'if' statement will make sure 270 is returned    
    x_vector = vec_azi[0] 
    if x_vector < 0:
        return 360 - angle2vecs(vec_azi, vec_N)
    else:
        return angle2vecs(vec_azi, vec_N)
        
def tilt(poly):
    """Tilt of a polygon poly"""
    N = len(poly) - 1
    vec = unit_normal(poly[0], poly[1], poly[N])
    vec_alt = np.array([vec[0], vec[1], vec[2]])
    vec_z = np.array([0,0,1])
    # return (90 - angle2vecs(vec_alt, vec_z)) # update by Santosh
    return (angle2vecs(vec_alt, vec_z))
    


# Copyright (c) 2012 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

"""do just walls in eplusinterface"""

from idfreader import idfreader


iddfile = "../iddfiles/Energy+V7_2_0.idd"
fname = "../idffiles/V_7_2/box_tiltrot.idf"
 
bunchdt, data, commdct = idfreader(fname, iddfile)
surfaces = bunchdt['BUILDINGSURFACE:DETAILED'.upper()]
# surfaces = bunchdt['FenestrationSurface:Detailed'.upper()]

wall = surfaces[0]
# n_vertices = 'Number_of_Vertices'
# n_vertices_i = wall.objls.index(n_vertices)


for surface in surfaces:
    name = surface.Name
    area = surface.area
    # height = surface.height
    # width = surface.width
    azimuth = surface.azimuth
    tilt = surface.tilt
    # coords = surface.coords
    # print name, area, height, width, azimuth, tilt
    # print name, azimuth
    print name, area, azimuth, tilt





# Copyright (c) 2012, 2013 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

"""function to build the airloop"""

def makezoneequipconnections(idf, loopname, zname):
    """make a ZoneHVAC:EquipmentConnections for zname"""
    pass



# EPlusInterface (EPI) - An interface for EnergyPlus
# Copyright (C) 2004 Santosh Philip

# This file is part of EPlusInterface.
# 
# EPlusInterface is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# EPlusInterface is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with EPlusInterface; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 


# Santosh Philip, the author of EPlusInterface, can be contacted at the following email address:
# santosh_philip AT yahoo DOT com
# Please send all bug reports, enhancement proposals, questions and comments to that address.
# 
# VERSION: 0.001

import mylib1,mylib2,mylib3,string

class eplusdict:
	"""
	use this when you want to use a dictionary
	but want to maintain the order of the items.
	There is a alist that keeps the order
	"""
	adict={}
	alist=[]
	def __init__(self,dct=None,lst=None):
		if dct==None:
			self.adict={}
		else:
			self.adict=dct
		if lst==None:
			self.alist=[]
		else:
			self.alist=lst
	
	def add(self,name,content):
		"""
		you can replace an existing value with this.
		Dictionary values are entered as lists:
		{key1:[val1,val2,val3],key2:[val4,val8]}
		if a key already exists, the value is added 
		to the list rather than replacing the existing list
		"""
		if not self.adict.has_key(name):
			self.adict[name]=[]
		self.adict[name].append(content)
		self.alist.append(name)
	
	def remove(self,name):
		del self.adict[name]
		self.alist.remove(name)
		
# 	def update(self,name,content):
# 		if not(self.adict.has_key(name)):
# 			raise ValueError,'%s does not exists in the ordereddict' %(`name`,)
# 		self.adict[name]=content
	
	
class ordereddict:
	"""
	use this when you want to use a dictionary
	but want to maintain the order of the items.
	There is a alist that keeps the order
	"""
	adict={}
	alist=[]
	def __init__(self,dct=None,lst=None):
		if dct==None:
			self.adict={}
		else:
			self.adict=dct
		if lst==None:
			self.alist=[]
		else:
			self.alist=lst
	
	def add(self,name,content,warn=1):
		"""
		you can replace an existing value with this
		just keep warn=0
		"""
		if warn:
			if self.adict.has_key(name):
				raise ValueError,'%s already exists in the ordereddict' %(`name`,)
		self.adict[name]=content
		try:
			#in case you are replacing an existing value
			self.alist.remove(name)
		except:
			pass		

		self.alist.append(name)
	
	def remove(self,name):
		del self.adict[name]
		self.alist.remove(name)
		
	def update(self,name,content):
		if not(self.adict.has_key(name)):
			raise ValueError,'%s does not exists in the ordereddict' %(`name`,)
		self.adict[name]=content

	def reverse(self):
		self.alist.reverse()

class eplusdict1(ordereddict):
	"""
	first cycle of extracting the dictionary from the .idd file
	"""
	linesep=''
	def __init__(self,st):
		linesep=mylib3.getlinesep(st)
		self.linesep=linesep
		nocom=nocomment(st)
		iddst=nocom.nocom.strip()
		st=self.removecommasemi(iddst)
		
		#b=eplus1.ordereddict()
		while st.find(';')!=-1:
			(command,this,rest)=self.commandthisrest(st)
			st=rest
			self.add(command,this)
		
		self.reverse()
	
	def removecommasemi(self,iddst):
		"""
		need to change , and ; from the comment(\) line
		replace , with . and ; with :
		"""
		slashcomment='\\'
		comma=','
		dot='.'
		semi=';'
		colon=':'
		linesep=self.linesep
		ls=iddst.split(linesep)  
		for i in range(len(ls)):
			a=ls[i]
			if a.find(slashcomment)!=-1:
				als=a.split(slashcomment)
				tmp=als[-1]
				als[-1]=als[-1].replace(comma,dot)
				als[-1]=als[-1].replace(semi,colon)
				ls[i]=string.join(als,slashcomment)
		
		iddst=string.join(ls,linesep)
		return iddst

	def commandthisrest(self,st):
		"""
		st = 
		\gumby
		Lead Input;
		
		Simulation Data;
		
		\group Simulation Parameters
		
		VERSION,
			  \unique-object
		  A1 ; \field Version Identifier
			  \required-field
		
		BUILDING,
			   \unique-object
			   \required-object
			   \min-fields 6
		  A1 , \field Building Name
			   \required-field
			   \default NONE
		  N1 , \field North Axis
			   \note degrees from true North
			   \units deg
			   \type real
			   \default 0.0
			   ...
			   ...
		============================
		commandthisrest works backwords thru st
		In this case
		command = BUILDING
		this = 
		BUILDING,
			   \unique-object
			   \required-object
			   \min-fields 6
		  A1 , \field Building Name
			   \required-field
			   \default NONE
		  N1 , \field North Axis
			   \note degrees from true North
			   \units deg
			   \type real
			   \default 0.0
			   ...
			   ...
		rest =
		\gumby
		Lead Input;
		
		Simulation Data;
		
		\group Simulation Parameters
		
		VERSION,
			  \unique-object
		  A1 ; \field Version Identifier
			  \required-field
		"""
		linesep=self.linesep
		ls1=st.split(';')
		ls2=ls1[-2].split(',')
		ls3=ls2[0].split(linesep)
		command=ls3[-1].strip()
		
		st=string.join(ls3[:-1],linesep)
		ls=ls1[:-2]
		ls.append(st)
		rest=string.join(ls,';')
		
		ls2[0]=ls3[-1]
		ls2.append(ls1[-1])
		this=string.join(ls2,',')
		return command,this,rest
	
class eplusdict2(eplusdict1):
	"""
	second cycle of extracting the dictionary from the .idd file
	"""
	def __init__(self,st):
		eplusdict1.__init__(self,st)

	def parsecommand1(self,key):
		linesep=self.linesep
		st=self.adict[key]
		st=st+linesep
		ls=st.split(',')
		for i in range(len(ls)):
			ls[i]=ls[i].split(linesep)
		
		for i in range(len(ls)-1):
			ls[i]=ls[i]+(ls[i+1][:-1])
			ls[i+1]=[ls[i+1][-1]]
		
		ls.pop()
		
		bb=ordereddict()
		for el in ls:
			key=el[0]
			bb.add(el[0].strip(),el[1:])
		
		return bb


class nocomment:
	"""
	strips the commentts form the energyplus file
	"""
	def __init__(self,st):
		self.nocom=self.stripcomments(st)
	
	def __repr__(self):
		return self.nocom
		
	def sp(self,a):
		return string.split(a,'!')
			
	def striplinecomments(self,aline):
		ls=self.sp(aline)
		#range(8)=[0, 1, 2, 3, 4, 5, 6, 7]
		#range(1,8,2)=[1, 3, 5, 7]
		popthis=range(1,len(ls),2)

		popthis.reverse()
		for i in popthis:
			ls.pop(i)
		return string.join(ls,'')
	
	def stripcomments(self,text):
		lsep=mylib3.getlinesep(text)
		ls=string.split(text,lsep)
		for i in range(len(ls)):
			ls[i]=self.striplinecomments(ls[i])
		return string.join(ls,lsep)
		
	def tofile(self,filename):
		mylib2.str2file(filename,`self`)


class nocomment1:
	"""
	strips the commentts form the energyplus file
	"""
	def __init__(self,st,c='!'):
		self.cm=c 
		self.nocom=self.stripcomments(st)
	
	def __repr__(self):
		return self.nocom
		
	def sp(self,a):
		return string.split(a,self.cm)
			
	def striplinecomments(self,aline):
		ls=self.sp(aline)
		#range(8)=[0, 1, 2, 3, 4, 5, 6, 7]
		#range(1,8,2)=[1, 3, 5, 7]
		popthis=range(1,len(ls),2)

		popthis.reverse()
		for i in popthis:
			ls.pop(i)
		return string.join(ls,'')
	
	def stripcomments(self,text):
		lsep=mylib3.getlinesep(text)
		ls=string.split(text,lsep)
		for i in range(len(ls)):
			ls[i]=self.striplinecomments(ls[i])
		return string.join(ls,lsep)
		
	def tofile(self,filename):
		mylib2.str2file(filename,`self`)


class parseidd:
	"""
	Complete parser for .idd file
	works well
	simply too slow
	takes 30 minutes on my G3
	I may never use.
	Instead use just in time parsing for each command as needed
	"""
	finallst=[]
	commentbin=[]
	wordbin=[]
	
	def __init__(self,st):
		self.finallst=[]
		self.commentbin=[]
		self.wordbin=[]
		sep=';'
		linesep=mylib3.getlinesep(st)
		st='dummy;'+linesep+st#the first command is skipped
		(comment,word,rest)=self.commentwordrest(st,sep,linesep)
		self.commentbin.append(comment)
		self.wordbin.append(word)
		while rest!=None:
			rest=self.commasemifinish(rest,linesep)
		
	
	def commentwordrest(self,st,sep,linesep):
		ls=st.split(sep)
		comment=ls[-1]
		rest=string.join(ls[:-1],sep)
		ls1=rest.split(linesep)
		lastline=ls1[-1]
		sentence=lastline
		if sentence.find(',')!=-1:
			ls2=sentence.split(',')
			word=ls2[-1].strip()
			#remove the word
			ls2.pop()
			lastline=string.join(ls2,',')
			ls1[-1]=lastline
		else:
			word=sentence.strip()
			ls1.pop()#remove the sentence/lastline
		rest=string.join(ls1,linesep)
		return (comment,word,rest)
	
	def commasemifinish(self,st,linesep):
		#no comma no semiis = finish
		if st.find(',')==-1 and st.find(';')==-1:
			self.finallst.reverse()
			return None
		else:
			ls1=st.split(',')
			ls2=st.split(';')
			comma=len(ls1[-1])
			semi=len(ls2[-1])
			if semi>comma:
				return self.docommastuff(st,linesep)
			else:
				return self.dosemistuff(st,linesep)
	
	def dofinish(self):
		print 'finished'
		
	
		
	def docommastuff(self,st,linesep):
		sep=','
		(comment,word,rest)=self.commentwordrest(st,sep,linesep)
		self.commentbin.append(comment)
		self.wordbin.append(word)
		print word
		#self.commasemifinish(rest,linesep)
		return rest
	
	
	def dosemistuff(self,st,linesep):
		self.doblock()
		sep=';'
		(comment,word,rest)=self.commentwordrest(st,sep,linesep)
		self.commentbin.append(comment)
		self.wordbin.append(word)
		print word
		#self.commasemifinish(rest,linesep)
		return rest
	
	def doblock(self):
		self.commentbin.reverse()
		self.wordbin.reverse()
		d=ordereddict()
		for i in range(len(self.wordbin)):
			d.add(self.wordbin[i],self.commentbin[i])
		self.commentbin=[]
		self.wordbin=[]
		self.finallst.append(d)
		del d
	



# Copyright (c) 2012 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

"""some helper files"""

from string import ascii_letters, digits

def onlylegalchar(name):
    """return only legal chars"""
    legalchar = ascii_letters + digits + ' '
    return ''.join([s for s in name[:] if s in legalchar])

def makefieldname(namefromidd):
    """made a field name that can be used by bunch"""
    newname = onlylegalchar(namefromidd)
    bunchname = newname.replace(' ', '_')
    return bunchname
    
def intinlist(lst):
    """test if int in list"""
    for item in lst:
        try:
            item = int(item)
            return True
        except ValueError, e:
            pass
    return False
    
def replaceint(fname, replacewith='%s'):
    """replace int in lst"""
    words = fname.split()
    for i, word in enumerate(words):
        try:
            word = int(word)
            words[i] = '%s'
        except ValueError, e:
            pass
    return ' '.join(words)

def cleaniddfield(acomm):
    """make all the keys lower case"""
    for key in acomm.keys():
        val = acomm[key]
        acomm[key.lower()] = val
    for key in acomm.keys():
        val = acomm[key]
        if key != key.lower():
            acomm.pop(key)
    return acomm
    
def cleancommdct(commdct):
    """make all keys in commdct lower case"""
    return [[cleaniddfield(fcomm) for fcomm in comm] for comm in commdct]    
    


# Copyright (c) 2012 Santosh Philip

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

"""draw plant and air loop
s_plantloop5.py + s_airloop2.py
removed arrow between demand and supply
has main with arguments

usage:
    python ex_loopdiagram.py fname.idf"""
# copy of s_airplantloop1.py

import pydot
import sys
import os
import getopt
# sys.path.append('../EPlusInputcode')
from EPlusInterfaceFunctions import readidf
import loops

help_message = '''
The help message goes here.
'''


class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg




def firstisnode(edge):
    if type(edge[0]) == tuple:
        return True
    else:
        return False

def secondisnode(edge):
    if type(edge[1]) == tuple:
        return True
    else:
        return False

def bothnodes(edge):
    if type(edge[0]) == tuple and type(edge[1]) == tuple:
        return True
    else:
        return False
    
def dropnodes(edges):
    """draw a graph without the nodes"""
    newedges = []
    added = False
    for edge in edges:
        if bothnodes(edge):
            newtup = (edge[0][0], edge[1][0])
            newedges.append(newtup)
            added = True
        elif firstisnode(edge):
            for edge1 in edges:
                if edge[0] == edge1[1]:
                    newtup = (edge1[0], edge[1])
                    try:
                        newedges.index(newtup)
                    except ValueError, e:
                        newedges.append(newtup)
                    added = True
        elif secondisnode(edge):
            for edge1 in edges:
                if edge[1] == edge1[0]:
                    newtup = (edge[0], edge1[1])
                    try:
                        newedges.index(newtup)
                    except ValueError, e:
                        newedges.append(newtup)
                    added = True
        # gets the hanging nodes - nodes with no connection
        if not added:
            if firstisnode(edge):
                newedges.append((edge[0][0], edge[1]))
            if secondisnode(edge):
                newedges.append((edge[0], edge[1][0]))
        added = False
    return newedges
    
    
def test_dropnodes():
    """py.test for dropnodes"""
    # test 1
    node = "node"
    (a,b,c,d,e,f,g,h,i) = (('a', node),'b',('c', node),'d',
        ('e', node),'f',('g', node),'h',('i', node))
    edges = [(a, b),
    (b, c),
    (c, d),
    (d, e),
    (e, f),
    (f, g),
    (g, h),
    (h, i),]
    theresult = [('a', 'b'), ('b', 'd'), ('d', 'f'), ('f', 'h'), ('h', 'i')]
    result = dropnodes(edges)
    assert result == theresult
    # test 2
    (a,b,c,d,e,f,g,h,i,j) = (('a', node),'b',('c', node),
        ('d', node),'e','f',('g', node),('h', node),'i',('j', node))
    edges = [(a, b),
    (b, c),
    (c, e),
    (e, g),
    (g, i),
    (i, j),
    (b, d),
    (d, f),
    (f, h),
    (h, i),]
    theresult = [('a', 'b'), ('b', 'e'), ('e', 'i'), ('i', 'j'), 
            ('b', 'f'), ('f', 'i')]
    result = dropnodes(edges)
    assert result == theresult
    
def makeanode(name):
    return pydot.Node(name, shape="plaintext", label=name)
    
def makeabranch(name):
    return pydot.Node(name, shape="box3d", label=name)

def makeendnode(name):
    return pydot.Node(name, shape="doubleoctagon", label=name, 
        style="filled", fillcolor="#e4e4e4")
    
def istuple(x):
    return type(x) == tuple

def nodetype(anode):
    """return the type of node"""
    try:
        return anode[1]
    except IndexError, e:
        return None


def edges2nodes(edges):
    """gather the nodes from the edges"""
    nodes = []
    for e1, e2 in edges:
        nodes.append(e1)
        nodes.append(e2)
    nodedict = dict([(n, None) for n in nodes])
    justnodes = nodedict.keys()
    justnodes.sort()
    return justnodes
    
def test_edges2nodes():
    """py.test for edges2nodes"""
    thedata = (([("a", "b"), ("b", "c"), ("c", "d")],
    ["a", "b", "c", "d"]), # edges, nodes
    )
    for edges, nodes in thedata:
        result = edges2nodes(edges)   
        assert result == nodes
        

def makediagram(edges):
    """make the diagram with the edges"""
    graph = pydot.Dot(graph_type='digraph')
    nodes = edges2nodes(edges)
    epnodes = [(node, 
        makeanode(node[0])) for node in nodes if nodetype(node)=="epnode"]
    endnodes = [(node, 
        makeendnode(node[0])) for node in nodes if nodetype(node)=="EndNode"]
    epbr = [(node, makeabranch(node)) for node in nodes if not istuple(node)]
    nodedict = dict(epnodes + epbr + endnodes)
    for value in nodedict.values():
        graph.add_node(value)
    for e1, e2 in edges:
        graph.add_edge(pydot.Edge(nodedict[e1], nodedict[e2]))
    return graph

def transpose2d(mtx):
    """Transpose a 2d matrix
       [
            [1,2,3],
            [4,5,6]
            ]
        becomes
        [
            [1,4],
            [2,5],
            [3,6]
            ]
    """
    trmtx = [[] for i in mtx[0]]
    for i in range(len(mtx)):
        for j in range(len(mtx[i])):
            trmtx[j].append(mtx[i][j])
    return trmtx
##   -------------------------    
##    from python cookbook 2nd edition page 162
    # map(mtx, zip(*arr))

def makebranchcomponents(data, commdct, anode="epnode"):
    """return the edges jointing the components of a branch"""
    alledges = []

    objkey = 'BRANCH'
    cnamefield = "Component %s Name"
    inletfield = "Component %s Inlet Node Name"
    outletfield = "Component %s Outlet Node Name"

    numobjects = len(data.dt[objkey])
    cnamefields = loops.repeatingfields(data, commdct, objkey, cnamefield)
    inletfields = loops.repeatingfields(data, commdct, objkey, inletfield)
    outletfields = loops.repeatingfields(data, commdct, objkey, outletfield)

    inlts = loops.extractfields(data, commdct, 
        objkey, [inletfields] * numobjects)
    cmps = loops.extractfields(data, commdct, 
        objkey, [cnamefields] * numobjects)
    otlts = loops.extractfields(data, commdct, 
        objkey, [outletfields] * numobjects)

    zipped = zip(inlts, cmps, otlts)
    tzipped = [transpose2d(item) for item in zipped]
    for i in range(len(data.dt[objkey])):
        tt = tzipped[i]
        # branchname = data.dt[objkey][i][1]
        edges = []
        for t0 in tt:
            edges = edges + [((t0[0], anode), t0[1]), (t0[1], (t0[2], anode))]
        alledges = alledges + edges
    return alledges

def makeairplantloop(data, commdct):
    """make the edges for the airloop and the plantloop"""
    anode = "epnode"
    endnode = "EndNode"

    # in plantloop get:
    #     demand inlet, outlet, branchlist
    #     supply inlet, outlet, branchlist
    plantloops = loops.plantloopfields(data, commdct)
    # splitters
    #     inlet
    #     outlet1
    #     outlet2
    splitters = loops.splitterfields(data, commdct)
    #     
    # mixer
    #     outlet
    #     inlet1
    #     inlet2

    mixers = loops.mixerfields(data, commdct)
    #     
    # supply barnchlist
    #     branch1 -> inlet, outlet
    #     branch2 -> inlet, outlet
    #     branch3 -> inlet, outlet
    #         
    # CONNET INLET OUTLETS
    edges = []

    # get all branches
    branchkey = "branch".upper()
    branches = data.dt[branchkey]
    branch_i_o = {}
    for br in branches:
        br_name = br[1]
        in_out = loops.branch_inlet_outlet(data, commdct, br_name)
        branch_i_o[br_name] = dict(zip(["inlet", "outlet"], in_out))
    # for br_name, in_out in branch_i_o.items():
    #     edges.append(((in_out["inlet"], anode), br_name))
    #     edges.append((br_name, (in_out["outlet"], anode)))

    # instead of doing the branch
    # do the content of the branch
    edges = makebranchcomponents(data, commdct)


    # connect splitter to nodes
    for splitter in splitters:
        # splitter_inlet = inletbranch.node
        splittername = splitter[0]
        inletbranchname = splitter[1] 
        splitter_inlet = branch_i_o[inletbranchname]["outlet"]
        # edges = splitter_inlet -> splittername
        edges.append(((splitter_inlet, anode), splittername))
        # splitter_outlets = ouletbranches.nodes
        outletbranchnames = [br for br in splitter[2:]]
        splitter_outlets = [branch_i_o[br]["inlet"] for br in outletbranchnames]
        # edges = [splittername -> outlet for outlet in splitter_outlets]
        moreedges = [(splittername, 
                            (outlet, anode)) for outlet in splitter_outlets]
        edges = edges + moreedges

    for mixer in mixers:
        # mixer_outlet = outletbranch.node
        mixername = mixer[0]
        outletbranchname = mixer[1] 
        mixer_outlet = branch_i_o[outletbranchname]["inlet"]
        # edges = mixername -> mixer_outlet
        edges.append((mixername, (mixer_outlet, anode)))
        # mixer_inlets = inletbranches.nodes
        inletbranchnames = [br for br in mixer[2:]]
        mixer_inlets = [branch_i_o[br]["outlet"] for br in inletbranchnames]
        # edges = [mixername -> inlet for inlet in mixer_inlets]
        moreedges = [((inlet, anode), mixername) for inlet in mixer_inlets]
        edges = edges + moreedges

    # connect demand and supply side
    # for plantloop in plantloops:
    #     supplyinlet = plantloop[1]
    #     supplyoutlet = plantloop[2]
    #     demandinlet = plantloop[4]
    #     demandoutlet = plantloop[5]
    #     # edges = [supplyoutlet -> demandinlet, demandoutlet -> supplyinlet]
    #     moreedges = [((supplyoutlet, endnode), (demandinlet, endnode)), 
    #         ((demandoutlet, endnode), (supplyinlet, endnode))]
    #     edges = edges + moreedges
    #     
    # -----------air loop stuff----------------------
    # from s_airloop2.py
    # Get the demand and supply nodes from 'airloophvac'
    # in airloophvac get:
    #   get branch, supplyinlet, supplyoutlet, demandinlet, demandoutlet
    objkey = "airloophvac".upper()
    fieldlists = [["Branch List Name",
        "Supply Side Inlet Node Name",
        "Demand Side Outlet Node Name",
        "Demand Side Inlet Node Names",
        "Supply Side Outlet Node Names"]] * loops.objectcount(data, objkey)
    airloophvacs = loops.extractfields(data, commdct, objkey, fieldlists)
    # airloophvac = airloophvacs[0]

    # in AirLoopHVAC:ZoneSplitter:
    #   get Name, inlet, all outlets
    objkey = "AirLoopHVAC:ZoneSplitter".upper()
    singlefields = ["Name", "Inlet Node Name"]
    fld = "Outlet %s Node Name"
    repeatfields = loops.repeatingfields(data, commdct, objkey, fld)
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    zonesplitters = loops.extractfields(data, commdct, objkey, fieldlists)

    # in AirLoopHVAC:SupplyPlenum:
    #   get Name, Zone Name, Zone Node Name, inlet, all outlets
    objkey = "AirLoopHVAC:SupplyPlenum".upper()
    singlefields = ["Name", "Zone Name", "Zone Node Name", "Inlet Node Name"]
    fld = "Outlet %s Node Name"
    repeatfields = loops.repeatingfields(data, commdct, objkey, fld)
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    supplyplenums = loops.extractfields(data, commdct, objkey, fieldlists)

    # in AirLoopHVAC:ZoneMixer:
    #   get Name, outlet, all inlets
    objkey = "AirLoopHVAC:ZoneMixer".upper()
    singlefields = ["Name", "Outlet Node Name"]
    fld = "Inlet %s Node Name"
    repeatfields = loops.repeatingfields(data, commdct, objkey, fld)
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    zonemixers = loops.extractfields(data, commdct, objkey, fieldlists)

    # in AirLoopHVAC:ReturnPlenum:
    #   get Name, Zone Name, Zone Node Name, outlet, all inlets
    objkey = "AirLoopHVAC:ReturnPlenum".upper()
    singlefields = ["Name", "Zone Name", "Zone Node Name", "Outlet Node Name"]
    fld = "Inlet %s Node Name"
    repeatfields = loops.repeatingfields(data, commdct, objkey, fld)
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    returnplenums = loops.extractfields(data, commdct, objkey, fieldlists)


    # connect room to each equip in equiplist
    # in ZoneHVAC:EquipmentConnections:
    #   get Name, equiplist, zoneairnode, returnnode
    objkey = "ZoneHVAC:EquipmentConnections".upper()
    singlefields = ["Zone Name", "Zone Conditioning Equipment List Name", 
        "Zone Air Node Name", "Zone Return Air Node Name"]
    repeatfields = []
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    equipconnections = loops.extractfields(data, commdct, objkey, fieldlists)
    # in ZoneHVAC:EquipmentList:
    #   get Name, all equiptype, all equipnames
    objkey = "ZoneHVAC:EquipmentList".upper()
    singlefields = ["Name", ]
    fieldlist = singlefields
    flds = ["Zone Equipment %s Object Type", "Zone Equipment %s Name"]
    repeatfields = loops.repeatingfields(data, commdct, objkey, flds)
    fieldlist = fieldlist + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    equiplists = loops.extractfields(data, commdct, objkey, fieldlists)
    equiplistdct = dict([(ep[0], ep[1:])  for ep in equiplists])
    for key, equips in equiplistdct.items():
        enames = [equips[i] for i in range(1, len(equips), 2)]
        equiplistdct[key] = enames
    # adistuunit -> room    
    # adistuunit <- VAVreheat 
    # airinlet -> VAVreheat
    # in ZoneHVAC:AirDistributionUnit:
    #   get Name, equiplist, zoneairnode, returnnode
    objkey = "ZoneHVAC:AirDistributionUnit".upper()
    singlefields = ["Name", "Air Terminal Object Type", "Air Terminal Name"]
    repeatfields = []
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    adistuunits = loops.extractfields(data, commdct, objkey, fieldlists)
    # code only for AirTerminal:SingleDuct:VAV:Reheat
    # get airinletnodes for vavreheats
    # in AirTerminal:SingleDuct:VAV:Reheat:
    #   get Name, airinletnode
    adistuinlets = loops.makeadistu_inlets(data, commdct)
    alladistu_comps = []
    for key in adistuinlets.keys():
        objkey = key.upper()
        singlefields = ["Name"] + adistuinlets[key]
        repeatfields = []
        fieldlist = singlefields + repeatfields
        fieldlists = [fieldlist] * loops.objectcount(data, objkey)
        adistu_components = loops.extractfields(data, commdct, objkey, fieldlists)
        alladistu_comps.append(adistu_components)

    # in AirTerminal:SingleDuct:Uncontrolled:
    #   get Name, airinletnode
    objkey = "AirTerminal:SingleDuct:Uncontrolled".upper()
    singlefields = ["Name", "Zone Supply Air Node Name"]
    repeatfields = []
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    uncontrolleds = loops.extractfields(data, commdct, objkey, fieldlists)


    #---------

    anode = "epnode"
    endnode = "EndNode"

    # edges = []

    # connect demand and supply side
    # for airloophvac in airloophvacs:
    #     supplyinlet = airloophvac[1]
    #     supplyoutlet = airloophvac[4]
    #     demandinlet = airloophvac[3]
    #     demandoutlet = airloophvac[2]
    #     # edges = [supplyoutlet -> demandinlet, demandoutlet -> supplyinlet]
    #     moreedges = [((supplyoutlet, endnode), (demandinlet, endnode)),
    #         ((demandoutlet, endnode), (supplyinlet, endnode))]
    #     edges = edges + moreedges

    # connect zonesplitter to nodes
    for zonesplitter in zonesplitters:
        name = zonesplitter[0]
        inlet = zonesplitter[1]
        outlets = zonesplitter[2:]
        edges.append(((inlet, anode), name))
        for outlet in outlets:
            edges.append((name, (outlet, anode)))

    # connect supplyplenum to nodes
    for supplyplenum in supplyplenums:
        name = supplyplenum[0]
        inlet = supplyplenum[3]
        outlets = supplyplenum[4:]
        edges.append(((inlet, anode), name))
        for outlet in outlets:
            edges.append((name, (outlet, anode)))

    # connect zonemixer to nodes
    for zonemixer in zonemixers:
        name = zonemixer[0]
        outlet = zonemixer[1]
        inlets = zonemixer[2:]
        edges.append((name, (outlet, anode)))
        for inlet in inlets:
            edges.append(((inlet, anode), name))

    # connect returnplenums to nodes
    for returnplenum in returnplenums:
        name = returnplenum[0]
        outlet = returnplenum[3]
        inlets = returnplenum[4:]
        edges.append((name, (outlet, anode)))
        for inlet in inlets:
            edges.append(((inlet, anode), name))

    # connect room to return node
    for equipconnection in equipconnections:
        zonename = equipconnection[0]
        returnnode = equipconnection[-1]
        edges.append((zonename, (returnnode, anode)))
    
    # connect equips to room
    for equipconnection in equipconnections:
        zonename = equipconnection[0]
        zequiplistname = equipconnection[1]
        for zequip in equiplistdct[zequiplistname]:
            edges.append((zequip, zonename))

    # adistuunit <- adistu_component 
    for adistuunit in adistuunits:
        unitname = adistuunit[0]
        compname = adistuunit[2]
        edges.append((compname, unitname))

    # airinlet -> adistu_component
    for adistu_comps in alladistu_comps:
        for adistu_comp in adistu_comps:
            name = adistu_comp[0]
            for airnode in adistu_comp[1:]:
                edges.append(((airnode, anode), name))

    # supplyairnode -> uncontrolled
    for uncontrolled in uncontrolleds:
        name = uncontrolled[0]
        airnode = uncontrolled[1]            
        edges.append(((airnode, anode), name))
                

    # edges = edges + moreedges    
    return edges


def getedges(fname, iddfile):
    """return the edges of the idf file fname"""
    data, commdct = readidf.readdatacommdct(fname, iddfile=iddfile)
    edges = makeairplantloop(data, commdct)
    return edges
    
def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        try:
            opts, args = getopt.getopt(argv[1:], "ho:v", ["help", "output="])
        except getopt.error, msg:
            raise Usage(msg)
    
        # option processing
        for option, value in opts:
            if option == "-v":
                verbose = True
            if option in ("-h", "--help"):
                raise Usage(help_message)
            if option in ("-o", "--output"):
                output = value
        
        if len(args) == 1:
            # iddfile = "../iddfiles/Energy+V6_0.idd"
            iddfile = "./resources/iddfiles/Energy+V7_0_0_036.idd"
            fname = args[0]
        elif len(args) == 2:
            iddfile = args[0]
            fname = args[1]
        # iddfile = "../iddfiles/Energy+V6_0.idd"
        print "readingfile"
        data, commdct = readidf.readdatacommdct(fname, iddfile=iddfile)
        print "constructing the loops"
        edges = makeairplantloop(data, commdct)
        # print edges
        print "making the diagram"
        g = makediagram(edges)
        dotname = '%s.dot' % (os.path.splitext(fname)[0])
        pngname = '%s.png' % (os.path.splitext(fname)[0])
        g.write(dotname)
        g.write_png(pngname)
    except Usage, err:
        print >> sys.stderr, sys.argv[0].split("/")[-1] + ": " + str(err.msg)
        print >> sys.stderr, "\t for help use --help"
        return 2


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) 2012 Tuan Tran

# This file is part of eppy.

# Eppy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Eppy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with eppy.  If not, see <http://www.gnu.org/licenses/>.

## {{{ http://code.activestate.com/recipes/578276/ (r1)
## modified by Tuan Tran trantuan@hawaii.edu at L+U, www.coolshadow.com

import numpy as np

# area of a polygon
def area(poly):
    if len(poly) < 3: # not a plane - no area
        return 0
    total = [0, 0, 0]
    N = len(poly)
    for i in range(N):
        vi1 = poly[i]
        vi2 = poly[(i+1) % N]
        prod = np.cross(vi1, vi2)
        total[0] += prod[0]
        total[1] += prod[1]
        total[2] += prod[2]
    result = np.dot(total, unit_normal(poly[0], poly[1], poly[2]))
    return abs(result/2)

# average height of a polygon
def height(poly):
    N = len(poly)
    h = 0.0
    for i in range(N):
        h += (poly[i][2])
    return h/N
        
#unit normal vector of plane defined by points a, b, and c
def unit_normal(a, b, c):
    x = np.linalg.det([[1,a[1],a[2]],[1,b[1],b[2]],[1,c[1],c[2]]])
    y = np.linalg.det([[a[0],1,a[2]],[b[0],1,b[2]],[c[0],1,c[2]]])
    z = np.linalg.det([[a[0],a[1],1],[b[0],b[1],1],[c[0],c[1],1]])
    magnitude = (x**2 + y**2 + z**2)**.5
    if magnitude < 0.00000001: mag = (0,0,0)
    else: mag = (x/magnitude, y/magnitude, z/magnitude)
    return mag
## end of http://code.activestate.com/recipes/578276/ }}}

